﻿using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Security;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.RegularExpressions;

namespace SECOM.Common.Services
{
    public interface ICertificateService
    {
        bool ValidateCertificate(X509Certificate2 cert, string rootCertificateTumbprint, out string errors);
        X509Certificate2 GetRootCertificate(X509Certificate2 cert);
        string GetBase64SingleLine(X509Certificate2 cert);
        X509Certificate2 GetX509CertFromBase64(string pem);
        string GetInstanceIdFromCert(X509Certificate2 cert);
    }

    public class CertificateService : ICertificateService
    {
        private readonly ILogger<CertificateService> _logger;

        public CertificateService(ILogger<CertificateService> logger)
        {
            _logger = logger;
        }

        public bool ValidateCertificate(X509Certificate2 cert, string rootCertificateThumbprint, out string errors)
        {
            errors = string.Empty;

            _logger.LogDebug("Checking certificate");

            if(string.IsNullOrEmpty(rootCertificateThumbprint))
            {
                rootCertificateThumbprint = GetRootCertificate(cert).Thumbprint;
            }
            
            X509Chain chain = new X509Chain();
            chain.ChainPolicy = new X509ChainPolicy()
            {
                RevocationFlag = X509RevocationFlag.EndCertificateOnly,
                RevocationMode = X509RevocationMode.NoCheck,
                UrlRetrievalTimeout = new TimeSpan(0, 0, 10),
                VerificationTime = DateTime.UtcNow
            };

            var chainBuilt = chain.Build(cert);
            _logger.LogDebug(string.Format("Chain building status: {0}", chainBuilt));

            if (chainBuilt)
            {
                if (chain.ChainElements
                    .Cast<X509ChainElement>()
                    .Any(x => x.Certificate.Thumbprint.ToUpper().Trim().CompareTo(rootCertificateThumbprint.ToUpper()) == 0))
                {
                    _logger.LogDebug("Certificate is valid");
                    return true;
                }
                else
                {
                    errors = "Invalid certificate, invalid issuer thumbprint";
                    _logger.LogInformation(errors);
                    return false;
                }
            }
            else
            {
                foreach (X509ChainStatus chainStatus in chain.ChainStatus)
                {
                    errors += string.Format("Chain error: {0} {1}", chainStatus.Status, chainStatus.StatusInformation);
                }

                _logger.LogInformation(errors);
                return false;
            }
        }

        public X509Certificate2 GetRootCertificate(X509Certificate2 cert)
        {
            X509Chain chain = new X509Chain();
            chain.ChainPolicy = new X509ChainPolicy()
            {
                RevocationFlag = X509RevocationFlag.EndCertificateOnly,
                RevocationMode = X509RevocationMode.NoCheck,
                UrlRetrievalTimeout = new TimeSpan(0, 0, 10),
                VerificationTime = DateTime.UtcNow
            };

            var chainBuilt = chain.Build(cert);
            if (chainBuilt)
            {
                return chain.ChainElements.Cast<X509ChainElement>().Last().Certificate;
            }
            return null;
        }

        public X509Certificate2 GetX509CertFromBase64(string publicCertBase64d)
        {
            string publicKey = "-----BEGIN CERTIFICATE-----" + Environment.NewLine;
            
            IEnumerable<string> data = GetChunks(publicCertBase64d, 64);
            foreach(var item in data)
            {
                publicKey += item + Environment.NewLine;
            }
            publicKey += "-----END CERTIFICATE-----" + Environment.NewLine;

            return new X509Certificate2(Encoding.UTF8.GetBytes(publicKey));
        }

        private IEnumerable<string> GetChunks(string str, int maxChunkSize)
        {
            for (int i = 0; i < str.Length; i += maxChunkSize)
                yield return str.Substring(i, Math.Min(maxChunkSize, str.Length - i));
        }

        public string GetBase64SingleLine(X509Certificate2 cert)
        {
            var sb = new StringBuilder();
            using (var sw = new StringWriter(sb))
            {
                var writer = new PemWriter(sw);
                writer.WriteObject(DotNetUtilities.FromX509Certificate(cert));
            }

            sb.Replace(Environment.NewLine, "");
            sb.Replace("-----BEGIN CERTIFICATE-----", "");
            sb.Replace("-----END CERTIFICATE-----", "");
            //Remove all remaining line breaks
            return sb.ToString().Replace(Environment.NewLine, "");
        }

        public string GetInstanceIdFromCert(X509Certificate2 cert)
        {
            // Extract data from certifcate
            var certData = cert.Subject.Split(',');

            foreach (var item in certData)
            {
                var parts = item.Split('=');
                if (parts != null && parts.Count() == 2)
                {
                    if (parts[0].Trim().Equals("OID.0.9.2342.19200300.100.1.1"))
                    {
                        return parts[1];
                    }
                    else if (parts[0].Trim().StartsWith("userId"))
                    {
                        return parts[1];
                    }
                }
            }

            return null;
        }

        private string GetCallerOrgMrnFromCert(X509Certificate2 cert)
        {
            // Extract data from certifcate
            var certData = cert.Subject.Split(',');

            var certDataDictionary = new Dictionary<string, string>();
            foreach (var item in certData)
            {
                var parts = item.Split('=');
                if (parts != null && parts.Count() == 2)
                {
                    if (parts[0].Trim() == "O")
                    {
                        return parts[1]; ;
                    }
                }
            }

            return null;
        }


        //public string GetThumbprint(Org.BouncyCastle.X509.X509Certificate input)
        //{
        //    X509Certificate2 cert = new X509Certificate2(input.GetEncoded());
        //    return cert.Thumbprint;
        //}
    }
}