﻿using Org.BouncyCastle.Utilities.Encoders;
using System;
using System.IO;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;

namespace SECOM.Common.Services
{
    public interface IEncryptionService
    {
        byte[] Encrypt(byte[] data, out byte[] key, out byte[] iv);
        byte[] Decrypt(byte[] data, byte[] key, byte[] iv);
        byte[] EncryptWithDerivedKey(byte[] data, X509Certificate2 senderPrivateCert, X509Certificate2 receiverPublicCert, byte[] iv);
        byte[] DecryptWithDerivedKey(byte[] encryptedData, X509Certificate2 receiverPrivateCert, X509Certificate2 senderPublicCert, byte[] iv);
    }

    public class EncryptionService : IEncryptionService
    {
        public EncryptionService(int keySize)
        {
            _keySize = keySize;
        }

        public int KeySize
        {
            get
            {
                return _keySize;
            }

            private set
            {
                if (value == _keySize) 
                    return;
                _keySize = value;
            }
        }
        private int _keySize { get; set; }
        public byte[] Encrypt(byte[] data, out byte[] key, out byte[] iv)
        {
            using (Aes aes = Aes.Create())
            {
                aes.KeySize = _keySize;
                aes.BlockSize = 128;
                aes.Padding = PaddingMode.PKCS7;
                aes.Mode = CipherMode.CBC;

                key = aes.Key;
                iv = aes.IV;

                ICryptoTransform encryptor = aes.CreateEncryptor(aes.Key, aes.IV);
                using (MemoryStream msEncryptor = new MemoryStream())
                {
                    using (CryptoStream csEncryptor = new CryptoStream(msEncryptor, encryptor, CryptoStreamMode.Write))
                    {
                        csEncryptor.Write(data, 0, data.Length);
                        csEncryptor.FlushFinalBlock();
                    }
                    return msEncryptor.ToArray();
                }
            }
        }

        public byte[] Decrypt(byte[] data, byte[] key, byte[] iv)
        {
            try
            {
                using (Aes aes = Aes.Create())
                {
                    aes.KeySize = key.Length * 8;
                    aes.BlockSize = 128;
                    aes.Padding = PaddingMode.PKCS7;
                    aes.Mode = CipherMode.CBC;
                    aes.Key = key;
                    aes.IV = iv;

                    using (MemoryStream msDecrypt = new MemoryStream())
                    {
                        using (CryptoStream csDecrypt = new CryptoStream(msDecrypt, aes.CreateDecryptor(), CryptoStreamMode.Write))
                        {
                            csDecrypt.Write(data, 0, data.Length);
                        }

                        return msDecrypt.ToArray();
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                return null;
            }
        }

        public byte[] EncryptWithDerivedKey(byte[] data, X509Certificate2 senderPrivateCert, X509Certificate2 receiverPublicCert, byte[] iv)
        {
            byte[] sessionKey = GetDerivedKey(senderPrivateCert, receiverPublicCert);

            using (Aes aes = new AesCryptoServiceProvider())
            {
                aes.KeySize = sessionKey.Length * 8;
                aes.BlockSize = 128;
                aes.Padding = PaddingMode.PKCS7;
                aes.Mode = CipherMode.CBC;
                aes.Key = sessionKey;
                aes.IV = iv;

                using (MemoryStream encrypted = new MemoryStream())
                using (CryptoStream cs = new CryptoStream(encrypted, aes.CreateEncryptor(), CryptoStreamMode.Write))
                {
                    cs.Write(data, 0, data.Length);
                    cs.FlushFinalBlock();
                    return encrypted.ToArray();
                }
            }
        }

        public byte[] DecryptWithDerivedKey(byte[] encryptedData, X509Certificate2 receiverPrivateCert, X509Certificate2 senderPublicCert, byte[] iv)
        {
            try
            {
                byte[] sessionKey = GetDerivedKey(receiverPrivateCert, senderPublicCert);
                System.Diagnostics.Debug.WriteLine(Convert.ToBase64String(sessionKey));
                using (Aes aes = new AesCryptoServiceProvider())
                {
                    aes.KeySize = sessionKey.Length * 8;
                    aes.BlockSize = 128;
                    aes.Padding = PaddingMode.PKCS7;
                    aes.Mode = CipherMode.CBC;

                    aes.Key = sessionKey;
                    aes.IV = iv;

                    using (MemoryStream plaintext = new MemoryStream())
                    {
                        using (CryptoStream cs = new CryptoStream(plaintext, aes.CreateDecryptor(), CryptoStreamMode.Write))
                        {
                            cs.Write(encryptedData, 0, encryptedData.Length);
                            cs.FlushFinalBlock();
                            return plaintext.ToArray();
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                throw;
                //return Encoding.UTF8.GetBytes($"Error in decrypt: {ex.Message}");
            }
        }

        public static byte[] GetDerivedKey(X509Certificate2 privateCertificate, X509Certificate2 publicCertificate)
        {
            byte[] derivedKey;

            using (var privateKey = (ECDsaCng)privateCertificate.GetECDsaPrivateKey())
            using (var publicKey = (ECDsaCng)publicCertificate.GetECDsaPublicKey())
            {
                var publicParams = publicKey.ExportParameters(false);

                using (var publicCng = ECDiffieHellman.Create(publicParams))
                using (var diffieHellman = new ECDiffieHellmanCng(privateKey.Key))
                {
                    derivedKey = diffieHellman.DeriveKeyMaterial(publicCng.PublicKey);
                }
            }

            return derivedKey;
        }
    }
}
