﻿using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Pkcs;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Utilities.Encoders;
using SECOM.Common.Models.Public;
using System;
using System.IO;
using System.Linq;
using System.Runtime.Serialization.Formatters.Binary;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.Json;

namespace SECOM.Common.Services
{
    public interface ISignatureService
    {
        bool Verify(DigitalSignatureValueObject signaruteObject, byte[] data, out string message);
        DigitalSignatureValueObject GenerateSignature(byte[] data, X509Certificate2 publicCertificate);

        DigitalSignatureValueObject GenerateSignature(object obj, X509Certificate2 publicCertificate);

        bool VerifyEnvelopeSignature(string envelopeSignature, byte[] signedData, string publicCertificate, out string message);
        bool VerifyEnvelopeSignature(object obj, string envelopeSignature, string publicCertificate, out string message);

    }

    public class SignatureService : ISignatureService
    {
        private readonly ILogger<SignatureService> _logger;
        private readonly ICertificateService _certificateService;
        private JsonSerializerOptions _serializeOptions;

        public SignatureService(ILogger<SignatureService> logger,
            ICertificateService certificateService)
        {
            _logger = logger;
            _certificateService = certificateService;
            _serializeOptions = new JsonSerializerOptions
            {
                PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
            };
        }

        public bool Verify(DigitalSignatureValueObject signaruteObject, byte[] data, out string message)
        {
            var isValid = false;

            var signedPublicKeyString = $"-----BEGIN CERTIFICATE-----\r\n{signaruteObject.PublicCertificate}\r\n-----END CERTIFICATE-----\r\n";
            Org.BouncyCastle.X509.X509Certificate signedPublicKey = null;
            using (var reader = new StringReader(signedPublicKeyString))
                signedPublicKey = (Org.BouncyCastle.X509.X509Certificate)new PemReader(reader).ReadObject();

            //ISigner signer = SignerUtilities.GetSigner(signedPublicKey.SigAlgName);
            ISigner signer = SignerUtilities.GetSigner("SHA-256withECDSA");
            signer.Init(false, signedPublicKey.GetPublicKey());
            signer.BlockUpdate(data, 0, data.Length);

            if (signer.VerifySignature(Hex.Decode(signaruteObject.DigitalSignature)))
            {
                message = "Signature validation ok";
                isValid = true;
            }
            else
            {
                message = "Invalid signature";
                isValid = false;
            }

            return isValid;
        }

        public DigitalSignatureValueObject GenerateSignature(object obj, X509Certificate2 publicCertificate)
        {
            // Load private key
            byte[] pkcs12Bytes = publicCertificate.Export(X509ContentType.Pkcs12, "dummy");
            Pkcs12Store pkcs12 = new Pkcs12StoreBuilder().Build();
            pkcs12.Load(new MemoryStream(pkcs12Bytes, false), "dummy".ToCharArray());

            ECPrivateKeyParameters privateKey = null;
            foreach (string alias in pkcs12.Aliases)
            {
                if (pkcs12.IsKeyEntry(alias))
                {
                    privateKey = (ECPrivateKeyParameters)pkcs12.GetKey(alias).Key;
                    break;
                }
            }

            //Build data
            var signatureDataBuilder = new Utils.CustomSerializer(obj.GetType());
            var signaturString = signatureDataBuilder.Serialize(obj);
            File.WriteAllText($@"c:\SECOM_Files\Debug\Signature-{obj.GetType()}_{Guid.NewGuid():n}.txt", signaturString);
            byte[] data = Encoding.UTF8.GetBytes(signaturString);

            // Create signature
            //ISigner signer = SignerUtilities.GetSigner(publicCertificate.SignatureAlgorithm.Value);
            ISigner signer = SignerUtilities.GetSigner("SHA-256withECDSA");
            signer.Init(true, privateKey);
            signer.BlockUpdate(data, 0, data.Length);
            var signature = signer.GenerateSignature();

            var rootCert = _certificateService.GetRootCertificate(publicCertificate);
            if (rootCert == null)
                _logger.LogWarning("Failed to get root cert from public cert");

            return new DigitalSignatureValueObject
            {
                DigitalSignature = Hex.ToHexString(signature),
                PublicCertificate = _certificateService.GetBase64SingleLine(publicCertificate),
                PublicRootCertificateThumbprint = rootCert?.Thumbprint
            };
        }

        public DigitalSignatureValueObject GenerateSignature(byte[] data, X509Certificate2 publicCertificate)
        {
            // Load private key
            byte[] pkcs12Bytes = publicCertificate.Export(X509ContentType.Pkcs12, "dummy");
            Pkcs12Store pkcs12 = new Pkcs12StoreBuilder().Build();
            pkcs12.Load(new MemoryStream(pkcs12Bytes, false), "dummy".ToCharArray());

            ECPrivateKeyParameters privateKey = null;
            foreach (string alias in pkcs12.Aliases)
            {
                if (pkcs12.IsKeyEntry(alias))
                {
                    privateKey = (ECPrivateKeyParameters)pkcs12.GetKey(alias).Key;
                    break;
                }
            }

            // Create signature
            //ISigner signer = SignerUtilities.GetSigner(publicCertificate.SignatureAlgorithm.Value);
            ISigner signer = SignerUtilities.GetSigner("SHA-256withECDSA");
            signer.Init(true, privateKey);
            signer.BlockUpdate(data, 0, data.Length);
            var signature = signer.GenerateSignature();

            return new DigitalSignatureValueObject
            {
                DigitalSignature = Hex.ToHexString(signature),
                PublicCertificate = _certificateService.GetBase64SingleLine(publicCertificate),
                PublicRootCertificateThumbprint = _certificateService.GetRootCertificate(publicCertificate).Thumbprint
            };
        }

        public bool VerifyEnvelopeSignature(object obj, string envelopeSignature, string publicCertificate, out string message)
        {
            var isValid = false;

            var signedPublicKeyString = $"-----BEGIN CERTIFICATE-----\r\n{publicCertificate}\r\n-----END CERTIFICATE-----\r\n";
            Org.BouncyCastle.X509.X509Certificate signedPublicKey = null;
            using (var reader = new StringReader(signedPublicKeyString))
                signedPublicKey = (Org.BouncyCastle.X509.X509Certificate)new PemReader(reader).ReadObject();


            //Build data
            var signatureDataBuilder = new Utils.CustomSerializer(obj.GetType());
            var signaturString = signatureDataBuilder.Serialize(obj);
            byte[] signedData = Encoding.UTF8.GetBytes(signaturString);

            ISigner signer = SignerUtilities.GetSigner("SHA-256withECDSA");
            signer.Init(false, signedPublicKey.GetPublicKey());
            signer.BlockUpdate(signedData, 0, signedData.Length);

            if (signer.VerifySignature(Hex.Decode(envelopeSignature)))
            {
                message = "Signature validation ok";
                isValid = true;
            }
            else
            {
                message = "Invalid signature";
                isValid = false;
            }

            return isValid;
        }

        public bool VerifyEnvelopeSignature(string envelopeSignature, byte[] signedData, string publicCertificate, out string message)
        {
            var isValid = false;

            var signedPublicKeyString = $"-----BEGIN CERTIFICATE-----\r\n{publicCertificate}\r\n-----END CERTIFICATE-----\r\n";
            Org.BouncyCastle.X509.X509Certificate signedPublicKey = null;
            using (var reader = new StringReader(signedPublicKeyString))
                signedPublicKey = (Org.BouncyCastle.X509.X509Certificate)new PemReader(reader).ReadObject();

            //ISigner signer = SignerUtilities.GetSigner(signedPublicKey.SigAlgName);
            ISigner signer = SignerUtilities.GetSigner("SHA-256withECDSA");
            signer.Init(false, signedPublicKey.GetPublicKey());
            signer.BlockUpdate(signedData, 0, signedData.Length);

            if (signer.VerifySignature(Hex.Decode(envelopeSignature)))
            {
                message = "Signature validation ok";
                isValid = true;
            }
            else
            {
                message = "Invalid signature";
                isValid = false;
            }

            return isValid;

        }
    }

}