﻿using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Http;
using SECOM.Common.Services;
using SECOM.Module.Ui.Services;
using SECOM.Module.Ui.ViewModels;
using SECOM.Module.Ui.Views;
using SECOM.Module.Ui.Views.Dialogs;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.IO;
using System.IO.IsolatedStorage;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Threading.Tasks;
using System.Windows;

namespace SECOM.Module.Ui
{
    public partial class App : Application
    {
        public IServiceProvider ServiceProvider { get; private set; }

        public IConfiguration Configuration { get; private set; }

        protected override void OnStartup(StartupEventArgs e)
        {
            if (!Directory.Exists(@"c:\SECOM_Files"))
                Directory.CreateDirectory(@"c:\SECOM_Files");
            if (!Directory.Exists(@"c:\SECOM_Files\Debug"))
                Directory.CreateDirectory(@"c:\SECOM_Files\Debug");

            Configuration = GetConfig();

            var serviceCollection = new ServiceCollection();
            ConfigureServices(serviceCollection);

            ServiceProvider = serviceCollection.BuildServiceProvider();
            LoadSettings();

            var mainWindow = ServiceProvider.GetRequiredService<MainWindow>();

            mainWindow.Show();
        }

        protected override void OnExit(ExitEventArgs e)
        {
            try
            {
                var settingsService = ServiceProvider.GetRequiredService<SettingsService>();

                IsolatedStorageFile isolatedStorage = IsolatedStorageFile.GetUserStoreForAssembly();
                using (var srWriter = new StreamWriter(new IsolatedStorageFileStream("StmModule", FileMode.Create, isolatedStorage)))
                {
                    srWriter.WriteLine(settingsService.Environment);
                    srWriter.WriteLine(settingsService.InstanceDisplayName);
                    srWriter.WriteLine(settingsService.InstanceName);
                    srWriter.WriteLine(settingsService.CertificateFile);
                    srWriter.WriteLine(settingsService.CertificatePassword);
                    srWriter.WriteLine(settingsService.EnvelopeCertificateFile);
                    srWriter.WriteLine(settingsService.EnvelopeCertificatePassword);
                    srWriter.Flush();
                }
            }
            catch (System.Security.SecurityException sx)
            {
                MessageBox.Show(sx.Message);
                throw;
            }
        }

        private void ConfigureServices(IServiceCollection services)
        {
            services.AddHttpClient("SecomPrivateService", c => GetDefaultHttpClient(c))
                .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());

            services.AddTransient<ISecomService, SecomService>();
            services.AddTransient<IEncryptionService>(p => new EncryptionService(256));
            services.AddTransient<ICertificateService, CertificateService>();
            services.AddTransient<ISignatureService, SignatureService>();
            services.AddTransient<ICompressionService, CompressionService>();
            services.AddTransient<ISentMessageService, SentMessageService>();


            services.AddSingleton<SettingsService>();
            services.AddSingleton<NotificationService>();
            services.AddSingleton(Configuration);

            services.AddTransient<MainViewModel>();
            services.AddTransient<MainWindow>();
            services.AddTransient<SettingsView>();
            services.AddTransient<PublishedMessageView>();
            services.AddTransient<UploadedMessageView>();
            services.AddTransient<FindServicesViewModel>();
            services.AddTransient<NewMessageDialogViewModel>();
            services.AddTransient<SendAckDialogViewModel>();

            services.AddTransient<SettingsViewModel>();
            services.AddTransient<PublishedMessageViewModel>();
            services.AddTransient<UploadedMessageViewModel>();
            services.AddTransient<FindServicesView>();
            services.AddTransient<NewMessageDialog>();
            services.AddTransient<SendAckDialog>();

            services.AddTransient<GetDialog>();
            services.AddTransient<GetViewModel>();

            services.AddTransient<GetSummaryDialog>();
            services.AddTransient<GetSummaryViewModel>();
        }

        private IConfiguration GetConfig()
        {
            var envVariable = Environment.GetEnvironmentVariable("DOTNET_ENVIRONMENT");

            var config = new ConfigurationBuilder();
            config.SetBasePath(Directory.GetCurrentDirectory());
            config.AddJsonFile("appsettings.json", optional: false, reloadOnChange: true);
            if (File.Exists($"appsettings.{envVariable}.json"))
                config.AddJsonFile($"appsettings.{envVariable}.json", optional: false, reloadOnChange: true);

            return config.Build();
        }

        private HttpClient GetDefaultHttpClient(HttpClient c)
        {
            c.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
            c.DefaultRequestHeaders.CacheControl = new CacheControlHeaderValue
            {
                NoCache = true,
                NoStore = true,
                MaxAge = new TimeSpan(0),
                MustRevalidate = true
            };

            return c;
        }

        private static HttpMessageHandler GetDefaultHttpMessageHandler()
        {
            return new HttpClientHandler
            {
                UseDefaultCredentials = true,
                ServerCertificateCustomValidationCallback = HttpClientHandler.DangerousAcceptAnyServerCertificateValidator,
                SslProtocols =
                            System.Security.Authentication.SslProtocols.Tls12 |
                            System.Security.Authentication.SslProtocols.Tls11 |
                            System.Security.Authentication.SslProtocols.Tls

            };
        }
        protected void LoadSettings()
        {
            try
            {
                var settingsService = ServiceProvider.GetRequiredService<SettingsService>();

                var isolatedStorage = IsolatedStorageFile.GetUserStoreForAssembly();
                using (var srReader = new StreamReader(new IsolatedStorageFileStream("StmModule", FileMode.OpenOrCreate, isolatedStorage)))
                {
                    if (srReader != null)
                    {
                        settingsService.Environment = srReader.ReadLine();
                        settingsService.InstanceDisplayName = srReader.ReadLine();
                        settingsService.InstanceName = srReader.ReadLine();
                        settingsService.CertificateFile = srReader.ReadLine();
                        settingsService.CertificatePassword = srReader.ReadLine();
                        settingsService.EnvelopeCertificateFile = srReader.ReadLine();
                        settingsService.EnvelopeCertificatePassword = srReader.ReadLine();
                    }
                }
            }
            catch (Exception)
            { }
        }
    }
}