﻿using AutoMapper;
using SECOM.Common.Integration.Models;
using SECOM.Common.Models.Private;
using System;
using System.Collections.Generic;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Text.Encodings.Web;
using System.Text.Json;
using System.Threading.Tasks;

namespace SECOM.Module.Ui.Services
{
    public interface ISecomService
    {
        Task<List<NotificationDto>> GetNotifications();
        Task<List<PublishedMessageDto>> GetPublishedMessages();
        Task<Guid> SavePublishedMessage(PublishedMessageDto message);
        Task DeletePublishedMessage(Guid dataReference);
        Task<List<UploadedMessageDto>> GetUploadedMessage();
        Task<EncryptionKeyDto> GetEncryptionKey(Guid transactionId);
        Task<List<FindServiceResponseObject>> FindService(string query, string geometry, int page, int pageSize);
        Task<CallServiceResponseDto> CallService(CallServiceRequestDto request);
        Task<Guid> SaveAcknowledgement(Common.Models.Public.AcknowledgementObject acknowledgement);
        Task<bool> SaveTemporaryKey(TemporaryKeyDTO keyObject);
    }

    public class SecomService : ISecomService
    {
        private readonly HttpClient _client;
        private readonly JsonSerializerOptions _serializeOptions;
        private readonly SettingsService _settingsService;

        public SecomService(IHttpClientFactory clientFactory,
            SettingsService settingsService)
        {
            _settingsService = settingsService;
            _client = clientFactory.CreateClient("SecomPrivateService");
            _serializeOptions = new JsonSerializerOptions
            {
                PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
            };
        }

        public async Task<List<NotificationDto>> GetNotifications()
        {
            var result = new List<NotificationDto>();
            using(var response = await _client.GetAsync($"{_settingsService.PrivateBaseUrl}/v1/Notification"))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var notifications = JsonSerializer.Deserialize<List<NotificationDto>>(jsonString, _serializeOptions);

                return notifications;
            }
        }

        public async Task<List<PublishedMessageDto>> GetPublishedMessages()
        {
            using (var response = await _client.GetAsync($"{_settingsService.PrivateBaseUrl}/v1/PublishedMessage"))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var messages = JsonSerializer.Deserialize<List<PublishedMessageDto>>(jsonString, _serializeOptions);

                return messages;
            }
        }

        public async Task DeletePublishedMessage(Guid dataReference)
        {
            using (var response = await _client.DeleteAsync($"{_settingsService.PrivateBaseUrl}/v1/PublishedMessage/{dataReference:n}"))
            {
                response.EnsureSuccessStatusCode();
            }
        }

        public async Task<List<UploadedMessageDto>> GetUploadedMessage()
        {
            using (var response = await _client.GetAsync($"{_settingsService.PrivateBaseUrl}/v1/UploadedMessage"))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var messages = JsonSerializer.Deserialize<List<UploadedMessageDto>>(jsonString, _serializeOptions);

                return messages;
            }
        }

        public async Task<EncryptionKeyDto> GetEncryptionKey(Guid transactionId)
        {
            using (var response = await _client.GetAsync($"{_settingsService.PrivateBaseUrl}/v1/EncryptionKey/{transactionId:n}"))
            {
                response.EnsureSuccessStatusCode();
                if (response.StatusCode == HttpStatusCode.NoContent)
                    return null;

                string jsonString = await response.Content.ReadAsStringAsync();
                var key = JsonSerializer.Deserialize<EncryptionKeyDto>(jsonString, _serializeOptions);

                return key;
            }
        }

        public async Task<List<FindServiceResponseObject>> FindService(string query, string geometry, int page, int pageSize)
        {
            using (var response = await _client.GetAsync($"{_settingsService.PrivateBaseUrl}/v1/ServiceRegistry/find?query={query}&page={page}&pagesize={pageSize}"))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var instances = JsonSerializer.Deserialize<List<FindServiceResponseObject>>(jsonString, _serializeOptions);

                return instances;
            }
        }

        public async Task<CallServiceResponseDto> CallService(CallServiceRequestDto request)
        {
            var stringContent = JsonSerializer.Serialize(request);
            var content = new StringContent(stringContent, Encoding.UTF8, "application/json");
            using (var response = await _client.PostAsync($"{_settingsService.PrivateBaseUrl}/v1/ServiceRegistry/call", content))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var result = JsonSerializer.Deserialize<CallServiceResponseDto>(jsonString, _serializeOptions);

                return result;
            }
        }

        public async Task<Guid> SaveAcknowledgement(Common.Models.Public.AcknowledgementObject acknowledgement)
        {
            var stringContent = JsonSerializer.Serialize(acknowledgement, _serializeOptions);
            var content = new StringContent(stringContent, Encoding.UTF8, "application/json");
            using (var response = await _client.PostAsync($"{_settingsService.PrivateBaseUrl}/v1/Acknowledgement", content))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var result = JsonSerializer.Deserialize<Guid>(jsonString, _serializeOptions);

                return result;
            }
        }

        public async Task<Guid> SavePublishedMessage(PublishedMessageDto message)
        {
            var stringContent = JsonSerializer.Serialize(message, _serializeOptions);
            var content = new StringContent(stringContent, Encoding.UTF8, "application/json");
            using (var response = await _client.PostAsync($"{_settingsService.PrivateBaseUrl}/v1/PublishedMessage", content))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var result = JsonSerializer.Deserialize<Guid>(jsonString, _serializeOptions);

                return result;
            }
        }

        public async Task<bool> SaveTemporaryKey(TemporaryKeyDTO message)
        {
            var stringContent = JsonSerializer.Serialize(message, _serializeOptions);
            var content = new StringContent(stringContent, Encoding.UTF8, "application/json");
            using (var response = await _client.PostAsync($"{_settingsService.PrivateBaseUrl}/v1/PublishedMessage/SaveTempKey", content))
            {
                response.EnsureSuccessStatusCode();
                string jsonString = await response.Content.ReadAsStringAsync();
                var result = JsonSerializer.Deserialize<bool>(jsonString, _serializeOptions);

                return result;
            }
        }
    }
}