﻿using SECOM.Common.Integration.Models;
using SECOM.Common.Models.Private;
using SECOM.Common.Models.Public;
using SECOM.Common.Services;
using SECOM.Module.Ui.Services;
using SECOM.Module.Ui.Utils;
using SECOM.Module.Ui.Views.Dialogs;
using Microsoft.Extensions.Configuration;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Security.Cryptography.X509Certificates;
using System.Text.Json;
using System.IO;
using System.Security;

namespace SECOM.Module.Ui.ViewModels
{
    public class PublishedMessageViewModel : ViewModelBase
    {
        private readonly SettingsService _settingsService;
        private readonly ISecomService _secomService;
        private readonly IServiceProvider _serviceProvider;
        private readonly ISignatureService _signatureService;
        private readonly IEncryptionService _encryptionService;
        private readonly ICompressionService _compressionService;
        private readonly ISentMessageService _sentMessageService;
        private readonly ICertificateService _certificateService;


        public PublishedMessageViewModel(SettingsService settingsService,
            ISecomService secomService,
            IConfiguration configuration,
            ISignatureService signatureService,
            IEncryptionService encryptionService,
            IServiceProvider serviceProvider,
            ICompressionService compressionService,
            ISentMessageService sentMessageService,
            ICertificateService certificateService)
        {
            _settingsService = settingsService;
            _secomService = secomService;
            _serviceProvider = serviceProvider;
            _signatureService = signatureService;
            _encryptionService = encryptionService;
            _compressionService = compressionService;
            _sentMessageService = sentMessageService;
            _certificateService = certificateService;

        }

        private PublishedMessageDto _selectedMessage;
        public PublishedMessageDto SelectedMessage
        {
            get
            {
                return _selectedMessage;
            }
            set
            {
                if (_selectedMessage == value)
                    return;

                _selectedMessage = value;

                OnPropertyChanged(() => SelectedMessage);
                OnPropertyChanged(() => SelectedMessageDataAsString);
                PublishMessageCommand.RaiseCanExecuteChanged();
                DeleteMessageCommand.RaiseCanExecuteChanged();
                ShowAclCommand.RaiseCanExecuteChanged();
                ShowSubscribersCommand.RaiseCanExecuteChanged();
            }
        }

        public string SelectedMessageDataAsString
        {
            get
            {
                if (SelectedMessage == null || SelectedMessage.Data != null)
                    return string.Empty;
                
                if (SelectedMessage.ContainerType == (int)ContainerTypeEnum.S100_EXCHANGESET)
                    return @$"Data is an Exchange set";

                return Encoding.UTF8.GetString(SelectedMessage.Data);
            }
        }


        private ObservableCollection<PublishedMessageDto> _messages = new ObservableCollection<PublishedMessageDto>();
        public ObservableCollection<PublishedMessageDto> Messages
        {
            get
            {
                return _messages;
            }
            set
            {
                if (_messages == value)
                    return;

                _messages = value;

                OnPropertyChanged(() => Messages);
            }
        }


        private DelegateCommand _loadPublishedMessagesCommand;
        public DelegateCommand LoadPublishedMessagesCommand
        {
            get
            {
                return _loadPublishedMessagesCommand ??
                    (_loadPublishedMessagesCommand = new DelegateCommand(ExecuteLoadPublishedMessagesCommand, CanExecuteLoadPublishedMessagesCommand));
            }
        }

        public bool CanExecuteLoadPublishedMessagesCommand(object parameter)
        {
            return true;
        }

        public async void ExecuteLoadPublishedMessagesCommand(object parameter)
        {
            Busy = true;
            BusyContent = "Loading published voyageplans from VIS";
            Messages = new ObservableCollection<PublishedMessageDto>(await _secomService.GetPublishedMessages());

            Busy = false;
        }


        private DelegateCommand _createNewMessageCommand;
        public DelegateCommand CreateNewMessageCommand
        {
            get
            {
                return _createNewMessageCommand ??
                    (_createNewMessageCommand = new DelegateCommand(ExecuteCreateNewMessageCommand, CanExecuteCreateNewMessageCommand));
            }
        }

        public bool CanExecuteCreateNewMessageCommand(object parameter)
        {
            return true;
        }

        public async void ExecuteCreateNewMessageCommand(object parameter)
        {
            //var dlg = new NewSTMMessageDialog();
            var dlg = (NewMessageDialog)_serviceProvider.GetService(typeof(NewMessageDialog));
            if(dlg.ShowDialog() == true)
            {
                var myCert = new X509Certificate2(_settingsService.CertificateFile, _settingsService.CertificatePassword, X509KeyStorageFlags.Exportable);
                byte[] data;

                var serializeOptions = new JsonSerializerOptions
                {
                    PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
                };

                data = File.ReadAllBytes(dlg.ViewModel.File);

                // Keep original data for signing
                byte[] originalData = new byte[data.Length];
                data.CopyTo(originalData, 0);

                // If compressionFlag, gzip data
                if (dlg.ViewModel.CompressionFlag)
                {
                    data = _compressionService.Compress(data);
                }

                // If data protection encrypt data
                if (dlg.ViewModel.DataProtection)
                {
                    data = _encryptionService.Encrypt(data, out var key, out var iv);
                    //Save tempKey to DB
                    //byte[] myCertExported = myCert.Export(X509ContentType.Pkcs12, _settingsService.CertificatePassword);
                    //X509Certificate2 copyCert = new X509Certificate2();
                    //copyCert.Import(myCertExported, _settingsService.CertificatePassword, X509KeyStorageFlags.Exportable);
                    System.Security.SecureString exportPwd = CreateSecureString();
                    byte[] certToExport = myCert.Export(X509ContentType.Pkcs12, exportPwd);

                    var saved = await _secomService.SaveTemporaryKey(new TemporaryKeyDTO 
                    {
                        Key = key,
                        IV = iv,
                        DataReference = dlg.ViewModel.TransactionId,
                        PrivateCertificate = certToExport,
                        Pwd = new System.Net.NetworkCredential(string.Empty, exportPwd).Password
                    });
                    if (saved)
                    {
                        MessageBox.Show($"Successfully saved temporary encryption key {dlg.ViewModel.TransactionId} to database");
                    }
                }
                var dataSig = _signatureService.GenerateSignature(originalData, myCert);
                var publishedMessage = new PublishedMessageDto
                {
                    Data = data,
                    ContainerType = dlg.ViewModel.ContainerType,
                    DataProductType = dlg.ViewModel.DataProductType,
                    DataReference = dlg.ViewModel.TransactionId,
                    ExchangeMetadataDto = new ExchangeMetadataDto
                    {
                        CompressionFlag = dlg.ViewModel.CompressionFlag,
                        DataProtection = dlg.ViewModel.DataProtection,
                        DigitalSignatureReference = dlg.ViewModel.DigitalSignatureReference,
                        ProtectionScheme = dlg.ViewModel.ProtectionScheme,
                        DigitalSignature = new DigitalSignatureDto
                        {
                            DigitalSignatureValue = dataSig.DigitalSignature,
                            PublicCertificate = dataSig.PublicCertificate,
                            PublicRootCertificateThumbprint = dataSig.PublicRootCertificateThumbprint
                        }
                    },
                    CreatedTime = DateTime.UtcNow,
                    //Geometry = @"LINESTRING(44 55, 45 56)",
                    Unlocode = dlg.ViewModel.Unlocode,
                    AckRequest = dlg.ViewModel.Acknowledgement
                };
                Messages.Add(publishedMessage);
                SelectedMessage = publishedMessage;
            }
        }

        private SecureString CreateSecureString()
        {
            var result = new SecureString();
            var arr = Guid.NewGuid();
            foreach(char c in arr.ToString().ToCharArray())
            {
                result.AppendChar(c);
            }
            return result;
        }

        private DelegateCommand _publishMessageCommand;
        public DelegateCommand PublishMessageCommand
        {
            get
            {
                return _publishMessageCommand ??
                    (_publishMessageCommand = new DelegateCommand(ExecutePublishMessageCommand, CanExecutePublishMessageCommand));
            }
        }

        public bool CanExecutePublishMessageCommand(object parameter)
        {
            return SelectedMessage != null;
        }

        public async void ExecutePublishMessageCommand(object parameter)
        {
            Busy = true;
            BusyContent = "Publishing message";
            await Task.Factory.StartNew(async () =>
            {
                try
                {
                    var result = await _secomService.SavePublishedMessage(SelectedMessage);
                    MessageBox.Show($"Successfully published message {result}.");
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString());
                }
                finally
                {
                    Busy = false;
                }
            });


            Busy = false;
        }

        private DelegateCommand _deleteMessageCommand;
        public DelegateCommand DeleteMessageCommand
        {
            get
            {
                return _deleteMessageCommand ??
                    (_deleteMessageCommand = new DelegateCommand(ExecuteDeleteMessageCommand, CanExecuteDeleteMessageCommand));
            }
        }

        public bool CanExecuteDeleteMessageCommand(object parameter)
        {
            return SelectedMessage != null;
        }

        public async void ExecuteDeleteMessageCommand(object parameter)
        {
            if (MessageBox.Show("Are you sure that you want to delete message?", "Delete message", MessageBoxButton.YesNo, MessageBoxImage.Warning) == MessageBoxResult.No)
                return;

            Busy = true;
            BusyContent = "Deleting message";
            try
            {
                await _secomService.DeletePublishedMessage(SelectedMessage.DataReference);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
            finally
            {
                Busy = false;
            }

            ExecuteLoadPublishedMessagesCommand(null);
        }

        private DelegateCommand _showAclCommand;
        public DelegateCommand ShowAclCommand
        {
            get
            {
                return _showAclCommand ??
                    (_showAclCommand = new DelegateCommand(ExecuteShowAclCommand, CanExecuteShowAclCommand));
            }
        }

        public bool CanExecuteShowAclCommand(object parameter)
        {
            return SelectedMessage != null;
        }

        public void ExecuteShowAclCommand(object parameter)
        {
            //var dlg = new AclDialog();
            //dlg.ViewModel = new VisAclViewModel();
            //dlg.ViewModel.LoadAcl(SelectedMessage.MessageID);
            //dlg.Show();
        }

        private DelegateCommand _showSubscribersCommand;
        public DelegateCommand ShowSubscribersCommand
        {
            get
            {
                return _showSubscribersCommand ??
                    (_showSubscribersCommand = new DelegateCommand(ExecuteShowSubscribersCommand, CanExecuteShowSubscribersCommand));
            }
        }

        public bool CanExecuteShowSubscribersCommand(object parameter)
        {
            return SelectedMessage != null;
        }

        public void ExecuteShowSubscribersCommand(object parameter)
        {
            //var dlg = new VisSubscribersDialog();
            //dlg.ViewModel.LoadSubscriptions(SelectedMessage.MessageID);
            //dlg.Show();
        }
    }
}