﻿using Org.BouncyCastle.OpenSsl;
using Org.BouncyCastle.Utilities.Encoders;
using SECOM.Common.Models.Private;
using SECOM.Common.Services;
using SECOM.Module.Ui.Extensions;
using SECOM.Module.Ui.Services;
using SECOM.Module.Ui.Utils;
using System;
using System.Collections.ObjectModel;
using System.Windows.Media;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using SECOM.Common.Models.Public;
using System.Text.Json;
using Org.BouncyCastle.Crypto.Parameters;
using System.Windows;
using SECOM.Module.Ui.Views.Dialogs;

namespace SECOM.Module.Ui.ViewModels
{
    public class UploadedMessageViewModel : ViewModelBase
    {
        private readonly ISecomService _secomService;
        private readonly ISignatureService _signatureService;
        private readonly IEncryptionService _encryptionService;
        private readonly ICompressionService _compressionService;
        private readonly SettingsService _settingsService;
        private readonly IServiceProvider _serviceProvider;
        private readonly ICertificateService _certificateService;

        public UploadedMessageViewModel(ISecomService secomService,
            ISignatureService signatureService,
            IEncryptionService encryptionService,
            ICompressionService compressionService,
            SettingsService settingsService,
            IServiceProvider serviceProvider,
            ICertificateService certificateService)
        { 
            _secomService = secomService;
            _signatureService = signatureService;
            _encryptionService = encryptionService;
            _compressionService = compressionService;
            _settingsService = settingsService;
            _serviceProvider = serviceProvider;
            _certificateService = certificateService;
        }


        private ExtendedUploadedObject _selectedMessage;
        public ExtendedUploadedObject SelectedMessage
        {
            get
            {
                return _selectedMessage;
            }
            set
            {
                if (_selectedMessage == value)
                    return;

                _selectedMessage = value;

                if (_selectedMessage == null)
                    return;

                ((DelegateCommand)SendAckCommand).RaiseCanExecuteChanged();
                OnPropertyChanged(() => SelectedMessage);
                OnPropertyChanged(() => SelectedMessageSignatureStatus);
            }
        }

        public Brush SelectedMessageSignatureStatus
        {
            get
            {
                if (SelectedMessage == null)
                    return Brushes.White;

                var signature = new Common.Models.Public.DigitalSignatureValueObject
                {
                    DigitalSignature = SelectedMessage.ExchangeMetadata.DigitalSignature.DigitalSignatureValue,
                    PublicCertificate = SelectedMessage.ExchangeMetadata.DigitalSignature.PublicCertificate,
                    PublicRootCertificateThumbprint = SelectedMessage.ExchangeMetadata.DigitalSignature.PublicRootCertificateThumbprint
                };

                var status = _signatureService.Verify(signature, SelectedMessage.Data, out string result);
                if (status == true)
                {
                    //_secomService.SetSignatureValidOk();
                    return Brushes.Green;
                }
                else
                {
                    return Brushes.Red;
                }
            }
        }

        private ObservableCollection<ExtendedUploadedObject> _messages = new ObservableCollection<ExtendedUploadedObject>();
        public ObservableCollection<ExtendedUploadedObject> Messages
        {
            get
            {
                return _messages;
            }
            set
            {
                if (_messages == value)
                    return;

                _messages = value;

                OnPropertyChanged(() => Messages);
            }
        }

        private ICommand _getMessagesCommand;
        public ICommand GetMessagesCommand
        {
            get
            {
                return _getMessagesCommand ??
                    (_getMessagesCommand = new DelegateCommand(ExecuteGetMessagesCommand, CanExecuteGetMessagesCommand));
            }
        }

        public bool CanExecuteGetMessagesCommand(object parameter)
        {
            return true;
        }

        public async void ExecuteGetMessagesCommand(object parameter)
        {
            ExtendedUploadedObject.SecomService = _secomService;
            ExtendedUploadedObject.CompressionService = _compressionService;
            ExtendedUploadedObject.EncryptionService = _encryptionService;
            ExtendedUploadedObject.SettingsService = _settingsService;
            ExtendedUploadedObject.SignatureService = _signatureService;

            var response = await _secomService.GetUploadedMessage();
            Messages = new ObservableCollection<ExtendedUploadedObject>(response.OrderBy(x => x.ReceivedTime).Select(x => new ExtendedUploadedObject(x)));
        }

        private ICommand _sendAckCommand;
        public ICommand SendAckCommand
        {
            get
            {
                return _sendAckCommand ??
                    (_sendAckCommand = new DelegateCommand(ExecuteSendAckCommand, CanExecuteSendAckCommand));
            }
        }

        private bool CanExecuteSendAckCommand(object parameter)
        {
            if (_selectedMessage != null)
                return !_selectedMessage.OperationalAckDelivered && (
                    _selectedMessage.AckRequest.HasValue ?
                    _selectedMessage.AckRequest.Value != (int)AckRequestEnum.No_ACK_requested && (
                    _selectedMessage.AckRequest.Value == (int)AckRequestEnum.Opened_ACK_requested ||
                    _selectedMessage.AckRequest.Value == (int)AckRequestEnum.Delivered_and_Opened_ACK_requested)
                    : false);
            else
                return false;
        }

        private async void ExecuteSendAckCommand(object parameter)
        {
            //if (_selectedMessage.OperationalAckDelivered)
            //{
            //    MessageBox.Show($"Operational acknowledgement for message {SelectedMessage.TransactionIdentifier} has already been sent!");
            //    return;
            //}
            var dlg = (SendAckDialog)_serviceProvider.GetService(typeof(SendAckDialog));
            dlg.ViewModel.TransactionId = _selectedMessage.TransactionIdentifier;
            if (dlg.ShowDialog() == true)
            {
                var serializeOptions = new JsonSerializerOptions
                {
                    PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
                };

                var envCert = new X509Certificate2(_settingsService.EnvelopeCertificateFile, _settingsService.EnvelopeCertificatePassword, X509KeyStorageFlags.Exportable);

                var ackData = new EnvelopeAckObject
                {
                    CreatedAt = DateTime.Now,
                    TransactionIdentifier = _selectedMessage.TransactionIdentifier,
                    EnvelopeCertificate = _certificateService.GetBase64SingleLine(envCert),
                    EnvelopeRootCertificateThumbprint = _certificateService.GetRootCertificate(envCert).Thumbprint,
                    EnvelopeSignatureTime = DateTime.UtcNow
                };

                ackData.AckType = (AckTypeEnum)dlg.ViewModel.AckType;
                if((int)dlg.ViewModel.NackType < 999)
                {
                    ackData.NackType = (NackTypeEnum)dlg.ViewModel.NackType;
                }

                var ackSignature = _signatureService.GenerateSignature(ackData, envCert);

                //Add signature to object
                var ackObject = new AcknowledgementObject
                {
                    Envelope = ackData,
                    DigitalSignature = ackSignature.DigitalSignature
                };

                var ackStringContent = JsonSerializer.Serialize(ackObject, serializeOptions);
                File.WriteAllText($@"c:\SECOM_Files\Debug\Ack_operational-{dlg.ViewModel.TransactionId:n}.txt", ackStringContent);

                var request = new CallServiceRequestDto
                {
                    EndpointMethod = SelectedMessage.AckEndpoint,
                    RequestType = "POST",
                    Body = ackStringContent,
                };

                //Send ack to SECOM ack receiver
                var response = await _secomService.CallService(request);
                if ((int)response.StatusCode < 200 || (int)response.StatusCode > 299)
                {
                    MessageBox.Show($"Unable to send acknowledgement. {response.StatusCode} - {response.Body}");
                }
                else
                {
                    //Set AckDelivered 
                    await _secomService.SaveAcknowledgement(ackObject);

                    MessageBox.Show($"Successfully sent acknowledgement to {SelectedMessage.AckEndpoint}.");
                }
            }
        }

        private DelegateCommand _viewCertificateCommand;
        public DelegateCommand ViewCertificateCommand
        {
            get
            {
                return _viewCertificateCommand ??
                    (_viewCertificateCommand = new DelegateCommand(ExecuteViewCertificateCommand, CanExecuteViewCertificateCommand));
            }
        }

        public bool CanExecuteViewCertificateCommand(object parameter)
        {
            return true;
        }

        public void ExecuteViewCertificateCommand(object parameter)
        {
            var signedPublicKeyString = $"-----BEGIN CERTIFICATE-----\r\n{SelectedMessage.ExchangeMetadata.DigitalSignature.PublicCertificate}\r\n-----END CERTIFICATE-----\r\n";
            Org.BouncyCastle.X509.X509Certificate cert = null;
            using (var reader = new StringReader(signedPublicKeyString))
                cert = (Org.BouncyCastle.X509.X509Certificate)new PemReader(reader).ReadObject();

            if (cert != null)
            {
                X509Certificate2UI.DisplayCertificate(new X509Certificate2(cert.GetEncoded()));
            }
        }
    }
}