﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using AutoMapper;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Logging;
using Org.BouncyCastle.Math.EC.Rfc7748;
using SECOM.Common.Models.Private;
using SECOM.Common.Models.Public;
using SECOM.DataAccess;

namespace SECOM.Service.Private.Rest.v1
{
    [Route("v1/[controller]")]
    [ApiController]
    public class PublishedMessageController : ControllerBase
    {
        private readonly ILogger _logger;
        private readonly SecomDbContext _dbContext;
        private readonly IMapper _mapper;

        public PublishedMessageController(ILogger<PublishedMessageController> logger,
            SecomDbContext dbContext,
            IMapper mapper,
            IHttpContextAccessor contextAccessor)
        {
            _logger = logger;
            _dbContext = dbContext;
            _mapper = mapper;

            _logger.LogDebug($"{GetType().FullName} accessed by user {contextAccessor.HttpContext.User.Identity.Name} from {contextAccessor.HttpContext.Connection.RemoteIpAddress}");

        }

        [HttpGet]
        public async Task<List<PublishedMessageDto>> Get()
        {
            return _mapper.Map<List<PublishedMessageDto>>(
                await _dbContext.PublishedMessage.ToListAsync());
        }

        [HttpPost]
        public async Task<ActionResult<Guid>> Post([FromBody]PublishedMessageDto message)
        {
            _logger.LogDebug($"Trying to save published message {message.DataReference} to database.");

            var check = _dbContext.PublishedMessage.Where(m => m.DataReference == message.DataReference).FirstOrDefault();
            if (check == null)
            {
                //var dbEntity = _mapper.Map<DataAccess.Entities.PublishedMessage>(message); 
                //map does not work for some reason, do manual mapping
                SECOM.DataAccess.Entities.PublishedMessage dbEntity = ConvertToEntity(message);
                await _dbContext.PublishedMessage.AddAsync(dbEntity);
                await _dbContext.SaveChangesAsync();
                _logger.LogDebug($"Successfully saved published message {message.DataReference} to database.");
                return dbEntity.DataReference;
            }
            else
            {
                var msg = $"Message with datareference {message.DataReference} already exists.";
                _logger.LogDebug(msg);
                return BadRequest(msg);
            }
        }

        [HttpPost]
        [Route("[action]")]
        public async Task<bool> SaveTempKey([FromBody] TemporaryKeyDTO keyObject)
        {
            _logger.LogDebug($"Trying to save temporary key {keyObject.DataReference} to database.");
            try
            {
                var dbEntity = _mapper.Map<DataAccess.Entities.TemporaryKey>(keyObject);
                await _dbContext.TemporaryKey.AddAsync(dbEntity);
                await _dbContext.SaveChangesAsync();
                _logger.LogDebug($"Successfully saved temporary key {keyObject.DataReference} to database.");
                return true;
            }
            catch(Exception ex)
            {
                _logger.LogError($"{ex}");
                return false;
            }
        }
        private DataAccess.Entities.PublishedMessage ConvertToEntity(PublishedMessageDto message)
        {
            return new DataAccess.Entities.PublishedMessage
            {
                CreatedTime = message.CreatedTime,
                Data = message.Data,
                DataReference = message.DataReference,
                ContainerType = message.ContainerType,
                DataProductTypeId = message.DataProductType,
                ExchangeMetadata = new DataAccess.Entities.ExchangeMetadata
                {
                    CompressionFlag = message.ExchangeMetadataDto.CompressionFlag,
                    DataProtection = message.ExchangeMetadataDto.DataProtection,
                    DigitalSignatureReference = message.ExchangeMetadataDto.DigitalSignatureReference,
                    ProtectionScheme = message.ExchangeMetadataDto.ProtectionScheme,
                    DigitalSignature = new DataAccess.Entities.DigitalSignature
                    {
                        DigitalSignatureValue = message.ExchangeMetadataDto.DigitalSignature.DigitalSignatureValue,
                        PublicCertificate = message.ExchangeMetadataDto.DigitalSignature.PublicCertificate,
                        PublicRootCertificateThumbprint = message.ExchangeMetadataDto.DigitalSignature.PublicRootCertificateThumbprint
                    }
                },
                Geometry = message.Geometry,
                Unlocode = message.Unlocode,
                AckRequest = message.AckRequest
            };
        }

        [HttpDelete]
        [Route("{dataReference}")]
        public async Task Delete([FromRoute]Guid dataReference)
        {
            var dbEntity = await _dbContext.PublishedMessage.FirstOrDefaultAsync(x => x.DataReference == dataReference);
            if (dbEntity == null)
                return;

            _dbContext.PublishedMessage.Remove(dbEntity);
            await _dbContext.SaveChangesAsync();
        }

    }
}