﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;
using AutoMapper;
using AutoMapper.Configuration;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Configuration;
using SECOM.Common.Integration.Models;
using SECOM.Common.Models.Private;
using SECOM.Common.Services;
using IConfiguration = Microsoft.Extensions.Configuration.IConfiguration;

namespace SECOM.Service.Private.Rest.v1
{
    [Route("v1/[controller]")]
    [ApiController]
    public class ServiceRegistryController : ControllerBase
    {
        private readonly ILogger _logger;
        private readonly IServiceRegistryService _serviceRegistryService;
        private readonly IConfiguration _configuration;

        public ServiceRegistryController(ILogger<ServiceRegistryController> logger,
            IServiceRegistryService serviceRegistryService,
            IHttpContextAccessor contextAccessor,
            IConfiguration configuration)
        {
            _logger = logger;
            _serviceRegistryService = serviceRegistryService;
            _configuration = configuration;

            _logger.LogDebug($"{GetType().FullName} accessed by user {contextAccessor.HttpContext.User.Identity.Name} from {contextAccessor.HttpContext.Connection.RemoteIpAddress}");
        }

        [HttpGet]
        [Route("Find")]
        public async Task<List<FindServiceResponseObject>> Find(string query, string geometry, int page, int pageSize)
        {
            return await _serviceRegistryService.FindService(query, geometry, page, pageSize);
        }

        [HttpPost]
        [Route("Call")]
        public async Task<CallServiceResponseDto> CallService([FromBody] CallServiceRequestDto callServiceObj)
        {
            var result = new CallServiceResponseDto();

            var path = _configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Path");
            var password = _configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Password");
            _logger.LogDebug($"Cert path: {path}");

            var sslCert = new X509Certificate2(path, password);
            var clientHandler = new HttpClientHandler();
            clientHandler.ClientCertificates.Add(sslCert);
            clientHandler.ClientCertificateOptions = ClientCertificateOption.Manual;
            clientHandler.ServerCertificateCustomValidationCallback = (message, cert, chain, errors) =>
            {
                return true;
            };

            //alternatively jack in a custom validator
            //clientHandler.ServerCertificateCustomValidationCallback = ValidateCertificate;

            using (var client = new HttpClient(clientHandler))
            {
                client.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
                client.DefaultRequestHeaders.CacheControl = new CacheControlHeaderValue
                {
                    NoCache = true,
                    NoStore = true,
                    MaxAge = new TimeSpan(0),
                    MustRevalidate = true
                };
                client.DefaultRequestHeaders.ConnectionClose = true;
                if (callServiceObj.Body != null)
                {
                    var uploadContent = new StringContent(callServiceObj.Body, Encoding.UTF8, "application/json");
                    using (HttpResponseMessage response = await client.PostAsync(callServiceObj.EndpointMethod, uploadContent))
                    {
                        result.StatusCode = response.StatusCode;
                        result.Body = await response.Content.ReadAsStringAsync();
                    }
                }
                else
                {
                    using(HttpResponseMessage response = await client.GetAsync(callServiceObj.EndpointMethod))
                    {
                        result.StatusCode = response.StatusCode;
                        result.Body = await response.Content.ReadAsStringAsync();
                    }
                }
            }
            _logger.LogInformation($"Returning result!");
            return result;
        }

        /// <summary>
        /// Custom certificate validator for validating the host certificate sent from receiving services through TLS
        /// </summary>
        /// <param name="request"></param>
        /// <param name="certificate"></param>
        /// <param name="certificateChain"></param>
        /// <param name="policy"></param>
        /// <returns></returns>
        private static bool ValidateCertificate(HttpRequestMessage request, X509Certificate2 certificate, X509Chain certificateChain, System.Net.Security.SslPolicyErrors policy)
        {
            var result = certificateChain.Build(certificate);
            if(result)
            {
                
                return true;
            }

            var errors = string.Empty;
            foreach (X509ChainStatus chainStatus in certificateChain.ChainStatus)
            {
                errors += string.Format("Chain error: {0} {1}", chainStatus.Status, chainStatus.StatusInformation);
            }

            return false;
        }

    }
}