﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.Json;
using System.Threading.Tasks;
using AutoMapper;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Logging;
using Org.BouncyCastle.Math.EC.Rfc7748;
using SECOM.Common.Models.Private;
using SECOM.Common.Models.Public;
using SECOM.DataAccess;

namespace SECOM.Service.Private.Rest.v1
{
    [Route("v1/[controller]")]
    [ApiController]
    public class UploadedMessageController : ControllerBase
    {
        private readonly ILogger _logger;
        private readonly SecomDbContext _dbContext;
        private readonly IMapper _mapper;
        private readonly Services.IAcknowledgementService _ackService;

        public UploadedMessageController(ILogger<UploadedMessageController> logger,
            SecomDbContext dbContext,
            IMapper mapper,
            IHttpContextAccessor contextAccessor, 
            Services.IAcknowledgementService acknowledgementService)
        {
            _logger = logger;
            _dbContext = dbContext;
            _mapper = mapper;
            _ackService = acknowledgementService;

            _logger.LogDebug($"{GetType().FullName} accessed by user {contextAccessor.HttpContext.User.Identity.Name} from {contextAccessor.HttpContext.Connection.RemoteIpAddress.MapToIPv4()}");
        }

        [HttpGet]
        public async Task<List<UploadedMessageDto>> Get()
        {
            var items = _mapper.Map<List<UploadedMessageDto>>(
                await _dbContext.UploadedMessage
                .Include(x => x.ExchangeMetadata)
                .Include(x => x.ExchangeMetadata.DigitalSignature).ToListAsync());

            foreach (var item in items)
            {
                if (!item.TechnicalAckDelivered && 
                    (item.AckRequest == (int)AckRequestEnum.Delivered_ACK_requested || item.AckRequest == (int)AckRequestEnum.Delivered_and_Opened_ACK_requested))
                {
                    var response = await _ackService.SendTechnicalAck(item);
                    if ((int)response.StatusCode >= 200 && (int)response.StatusCode < 300)
                    {
                        await SetAckDelivered(item.TransactionIdentifier);
                    }
                    else
                    {
                        //Skapa notifiering
                        var msg = response.Body.Split(":");
                        var notification = new DataAccess.Entities.Notification();
                        notification.Subject = $"Failed to send acknowledgement.";
                        notification.Body = $"Message id: {item.TransactionIdentifier} {Environment.NewLine}" +
                            $" Service response code: {response.StatusCode} {Environment.NewLine}. " +
                            $"Response message: {msg[1]}. {Environment.NewLine}" +
                            $"Receiving service: {item.AckEndpoint}";

                        notification.ReceivedAt = DateTime.Now;
                        notification.TransactionId = item.TransactionIdentifier;

                        await _dbContext.AddAsync(notification);
                        await _dbContext.SaveChangesAsync();
                    }
                }
            }

            return items;
        }

        [HttpGet]
        [Route("{dataReference}")]
        public async Task SetAckDelivered([FromRoute] Guid dataReference)
        {
            var dbEntity = await _dbContext.UploadedMessage.FirstOrDefaultAsync(x => x.TransactionIdentifier == dataReference);
            if (dbEntity == null)
                return;
            dbEntity.TechnicalAckDelivered = true;

            await _dbContext.SaveChangesAsync();
        }
    }
}