﻿using Microsoft.Extensions.Configuration;
using SECOM.Common.Models.Private;
using SECOM.Common.Services;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.Json;
using System.Threading.Tasks;
using IConfiguration = Microsoft.Extensions.Configuration.IConfiguration;

namespace SECOM.Service.Private.Services
{
    public interface IAcknowledgementService
    {
        Task<CallServiceResponseDto> SendTechnicalAck(UploadedMessageDto message);
    }
    public class AcknowledgementService : IAcknowledgementService
    {
        private readonly JsonSerializerOptions _serializeOptions;
        private readonly IConfiguration _configuration;
        private readonly ICertificateService _certificateService;
        private readonly ISignatureService _signatureService;

        public AcknowledgementService(IConfiguration configuration, ICertificateService certificateService, ISignatureService signatureService)
        {
            _configuration = configuration;
            _certificateService = certificateService;
            _signatureService = signatureService;
            _serializeOptions = new JsonSerializerOptions
            {
                PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
            };
        }
            
        public async Task<CallServiceResponseDto> SendTechnicalAck(UploadedMessageDto message)
        {
            var result = new CallServiceResponseDto();

            var path = _configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Path");
            var password = _configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Password");
            var sslCert = new X509Certificate2(path, password, X509KeyStorageFlags.Exportable);
            var clientHandler = new HttpClientHandler();
            clientHandler.ClientCertificates.Add(sslCert);
            clientHandler.ClientCertificateOptions = ClientCertificateOption.Manual;
            clientHandler.ServerCertificateCustomValidationCallback = (message, cert, chain, errors) =>
            {
                return true;
            };

            var publicCert = _certificateService.GetBase64SingleLine(sslCert);

            var ackObject = new AcknowledgementDto
            {
                Envelope = new EnvelopeAckDto
                {
                    CreatedAt = DateTime.Now,
                    EnvelopeCertificate = publicCert,
                    EnvelopeRootCertificateThumbprint = _certificateService.GetRootCertificate(sslCert).Thumbprint,
                    TransactionIdentifier = message.TransactionIdentifier,
                    Type = Common.Models.Public.AckTypeEnum.Delivered,
                    Reason = null,
                    EnvelopeSignatureTime = DateTime.UtcNow
                },
            };

            var signature = _signatureService.GenerateSignature(ackObject.Envelope, sslCert);
            ackObject.DigitalSignature = signature.DigitalSignature;
            
            using(var client = new HttpClient(clientHandler))
            {
                client.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
                client.DefaultRequestHeaders.CacheControl = new CacheControlHeaderValue
                {
                    NoCache = true,
                    NoStore = true,
                    MaxAge = new TimeSpan(0),
                    MustRevalidate = true
                };

                var ackStringContent = JsonSerializer.Serialize(ackObject, _serializeOptions);
                File.WriteAllText($@"c:\SECOM_Files\Debug\Ack_technical-{ackObject.Envelope.TransactionIdentifier:n}.txt", ackStringContent);
                var uploadContent = new StringContent(ackStringContent, Encoding.UTF8, "application/json");
                using (HttpResponseMessage response = await client.PostAsync(message.AckEndpoint, uploadContent))
                {
                    result.StatusCode = response.StatusCode;
                    result.Body = await response.Content.ReadAsStringAsync();
                }
            }

            return result;    
        }
    }
}
