using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication.Certificate;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.OpenApi.Models;
using SECOM.DataAccess;
using AutoMapper;
using SECOM.Common.Services;
using SECOM.Service.Private.Services;

namespace SECOM.Service.Private
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        public void ConfigureServices(IServiceCollection services)
        {
            services.AddHttpContextAccessor();
            var sqlConnectionString = Configuration.GetConnectionString("DataAccessPostgreSqlProvider");
            services.AddDbContext<SecomDbContext>(options =>
                    options.UseNpgsql(
                        sqlConnectionString,
                        b => b.MigrationsAssembly("SECOM.DataAccess")
                    )
                );

            services.AddHttpClient("IdentityRegistry", c => GetDefaultHttpClient(c, Configuration.GetValue<string>("ServiceEndpoints:IdentityRegistry")))
                .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());
            services.AddHttpClient("ServiceRegistry", c => GetDefaultHttpClient(c, Configuration.GetValue<string>("ServiceEndpoints:ServiceRegistry")))
                .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());

            services.AddTransient<IServiceRegistryService, ServiceRegistryService>();
            services.AddScoped<ICertificateService, CertificateService>();
            services.AddScoped<ISignatureService, SignatureService>();
            services.AddScoped<IAcknowledgementService, AcknowledgementService>();

            services.AddControllers();

            services.AddAutoMapper(typeof(Startup));

            services.AddMvc()
                .ConfigureApiBehaviorOptions(options =>
                {
                    options.SuppressModelStateInvalidFilter = true;
                });

            services.AddSwaggerGen(c =>
            {
                c.SwaggerDoc("v1", new OpenApiInfo
                {
                    Title = "SECOM Private Service interface",
                    Version = "v1",
                    Description = " ",
                    Contact = new OpenApiContact
                    {
                        Name = "Swedish Maritime Administration",
                        Url = new Uri("http://sjofartsverket.se")
                    },
                });
                c.IncludeXmlComments(Path.Combine(AppDomain.CurrentDomain.BaseDirectory, @"SECOM.Service.Private.xml"));
            });

        }

        public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
        {
            app.UseSwagger();
            app.UseSwaggerUI(c =>
            {
                c.SwaggerEndpoint("./swagger/v1/swagger.json", "SECOM Private Service interface");
                c.DocumentTitle = "SECOM Private Service interface";
                c.RoutePrefix = string.Empty;
            });

            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }

            app.UseRouting();

            app.UseEndpoints(endpoints =>
            {
                app.UseEndpoints(endpoints =>
                {
                    endpoints.MapControllers();
                });
            });

            //using (var scope = app.ApplicationServices.GetService<IServiceScopeFactory>().CreateScope())
            //{
            //    scope.ServiceProvider.GetRequiredService<SecomDbContext>().Database.Migrate();
            //}
        }

        private HttpClient GetDefaultHttpClient(HttpClient c, string baseUtl)
        {
            c.BaseAddress = new Uri(baseUtl);
            c.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
            c.DefaultRequestHeaders.CacheControl = new CacheControlHeaderValue
            {
                NoCache = true,
                NoStore = true,
                MaxAge = new TimeSpan(0),
                MustRevalidate = true
            };

            return c;
        }

        private HttpMessageHandler GetDefaultHttpMessageHandler()
        {
            var path = Configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Path");
            var password = Configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Password");

            var sslCert = new X509Certificate2(path, password);
            var httpHandler = new HttpClientHandler
            {
                ClientCertificateOptions = ClientCertificateOption.Manual,
                ServerCertificateCustomValidationCallback = HttpClientHandler.DangerousAcceptAnyServerCertificateValidator,
                SslProtocols =
                            System.Security.Authentication.SslProtocols.Tls12 |
                            System.Security.Authentication.SslProtocols.Tls11 |
                            System.Security.Authentication.SslProtocols.Tls

            };
            httpHandler.ClientCertificates.Add(sslCert);
            return httpHandler;
        }
    }
}
