﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using Microsoft.Extensions.Logging;
using SECOM.Common.Services;
using System.Net.Http;
using Microsoft.Extensions.Configuration;
using Microsoft.EntityFrameworkCore;
using SECOM.Common.Models.Public;
using System.Net;
using System.Text.Json;
using System.IO;

namespace SECOM.Service.Public.BackgroundTask
{
    public class EncryptionKeyWorker
    {
        private readonly ILogger _logger;
        private readonly IServiceRegistryService _serviceRegistryService;
        private readonly ICertificateService _certificateService;
        private readonly IConfiguration _configuration;
        private readonly ISignatureService _signatureService;
        private readonly IHttpClientFactory _clientFactory;
        private readonly IEncryptionService _encryptionService;
        private readonly ICallService _callService;

        public EncryptionKeyWorker(ILogger<EncryptionKeyWorker> logger,
            IServiceRegistryService serviceRegistryService,
            ICertificateService certificateService,
            IHttpClientFactory clientFactory,
            IConfiguration configuration,
            ISignatureService signatureService,
            IEncryptionService encryptionService,
            ICallService callService)
        {
            _logger = logger;
            _serviceRegistryService = serviceRegistryService;
            _certificateService = certificateService;
            _clientFactory = clientFactory;
            _configuration = configuration;
            _signatureService = signatureService;
            _encryptionService = encryptionService;
            _callService = callService;
            
        }
        public X509Certificate2 PublicCertificate { get; set; }
        public X509Certificate2 ClientCertificate { get; set; }
        public Guid DataReference { get; set; }

        public DataAccess.Entities.TemporaryKey StoredKey { get; set; }

        public string Url { get; set; }

        public async Task SendKey(CancellationToken cancellationToken)
        {
            var callerMrn = _certificateService.GetInstanceIdFromCert(ClientCertificate);
            var serviceInstance = await _serviceRegistryService.FindService($"instanceId:{callerMrn.Replace(":", "\\:")}", "", 0, 1000);

            if (serviceInstance == null || serviceInstance.Count == 0)
                return;

            if (string.IsNullOrEmpty(Url))
                Url = $"{serviceInstance.FirstOrDefault().EndpointUri}/v1/EncryptionKey";

            //For debugging localhost

            //if (System.Diagnostics.Debugger.IsAttached)
            //{
            //    url = $"https://localhost:5001/v1/EncryptionKey";
            //}

            var myCert = new X509Certificate2(StoredKey.PrivateCertificate, StoredKey.Pwd, X509KeyStorageFlags.Exportable);
            var protecedEncryptionKey =_encryptionService.EncryptWithDerivedKey(StoredKey.Key, myCert, PublicCertificate, StoredKey.IV);

            //Build encryptionKey json payload
            var encryptionKeyObject = new EncryptionKeyObject
            {
                Envelope = new EnvelopeKeyObject
                {
                    EncryptionKey = protecedEncryptionKey,
                    IV = StoredKey.IV,
                    TransactionIdentifier = DataReference,
                    DigitalSignatureValue = _signatureService.GenerateSignature(protecedEncryptionKey, myCert),
                    EnvelopeSignatureCertificate = _certificateService.GetBase64SingleLine(myCert),
                    EnvelopeRootCertificateThumbprint = _certificateService.GetRootCertificate(myCert).Thumbprint,
                    EnvelopeSignatureTime = DateTime.UtcNow,
                }
            };
            var envelopeSign = _signatureService.GenerateSignature(encryptionKeyObject.Envelope, myCert);
            encryptionKeyObject.EnvelopeSignature = envelopeSign.DigitalSignature;
            var serializeOptions = new JsonSerializerOptions
            {
                PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
            };
            var keyStringContent = JsonSerializer.Serialize(encryptionKeyObject, serializeOptions);
            var keyRequest = new CallServiceRequestDto
            {
                EndpointMethod = Url,
                RequestType = "POST",
                Body = keyStringContent
            };

            //send key
            var keyResponse = await _callService.DoCall(keyRequest);
            if ((int)keyResponse.StatusCode < 200 ||
                (int)keyResponse.StatusCode > 299)
            {
                _logger.LogError($"Unable to upload. {keyResponse.StatusCode} - {keyResponse.Body}");
            }

            File.WriteAllText($@"c:\SECOM_Files\Debug\EncryptionKey-{encryptionKeyObject.Envelope.TransactionIdentifier:n}.txt", keyStringContent);
        }
    }
}
