﻿
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.Logging;
using SECOM.Common.Models.Public;
using SECOM.Common.Services;
using SECOM.DataAccess;
using SECOM.DataAccess.Entities;
using System;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Threading.Tasks;

namespace SECOM.Service.Public.BackgroundTask
{
    public class UploadLinkWorker
    {
        private readonly ILogger _logger;
        private readonly IServiceRegistryService _serviceRegistryService;
        private readonly ICertificateService _certificateService;
        private readonly IHttpClientFactory _clientFactory;
        private readonly IConfiguration _configuration;

        public UploadLinkWorker(ILogger<UploadLinkWorker> logger,
            IServiceRegistryService serviceRegistryService,
            ICertificateService certificateService,
            IHttpClientFactory clientFactory,
            IConfiguration configuration)
        {
            _logger = logger;
            _serviceRegistryService = serviceRegistryService;
            _certificateService = certificateService;
            _clientFactory = clientFactory;
            _configuration = configuration;
        }

        public UploadLinkObject UploadLinkObject { get; set; }
        public X509Certificate2 ClientCertificate { get; set; }

        public string AckEndpoint { get; set; }

        public string Url { get; set; }

        public async Task GetLink(CancellationToken cancellationToken)
        {
            var callerMrn = _certificateService.GetInstanceIdFromCert(ClientCertificate);
            var serviceInstance = await _serviceRegistryService.FindService($"instanceId:{callerMrn.Replace(":", "\\:")}", "", 0, 1000);

            if (serviceInstance == null || serviceInstance.Count == 0)
                return;

            if(string.IsNullOrEmpty(Url))
                Url = $"{serviceInstance.FirstOrDefault().EndpointUri}/v1/object/link?TransactionIdentifier={UploadLinkObject.Envelope.TransactionIdentifier}";

            //For debugging use localhost
            //if (IsLocalHost)
            //{
            //    var urls = Environment.GetEnvironmentVariable("ASPNETCORE_URLS")?.Split(";");
            //    url = $"{urls[0]}/v1/object/link?TransactionIdentifier={UploadLinkObject.Envelope.TransactionIdentifier}";
            //}
            var client = _clientFactory.CreateClient("GetByLink");
            client.DefaultRequestHeaders.Accept.Clear();
            client.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/octet-stream"));

            _logger.LogInformation($"Fetching data from link {UploadLinkObject.Envelope.TransactionIdentifier} with size {UploadLinkObject.Envelope.Size} kB.");
            var response = await client.GetAsync(Url);
            response.EnsureSuccessStatusCode();

            var data = await response.Content.ReadAsByteArrayAsync();

            //For debugging on localhost
            //File.WriteAllBytes($@"c:\SECOM_Files\Debug\GetByLinkResponse-{UploadLinkObject.Envelope.TransactionIdentifier:n}.bin", data);
            //File.WriteAllText($@"c:\SECOM_Files\Debug\GetByLinkQuery.txt", Url);

            var uploadedMessage = new UploadedMessage
            {
                AckEndpoint = AckEndpoint,
                AckRequest = (int)UploadLinkObject.Envelope.AckRequest,
                Data = data,
                ContainerType = (int)UploadLinkObject.Envelope.ContainerType,
                DataProductType = (int)UploadLinkObject.Envelope.DataProductType,
                ExchangeMetadata = new SECOM.DataAccess.Entities.ExchangeMetadata
                {
                    CompressionFlag = UploadLinkObject.Envelope.ExchangeMetadata.CompressionFlag,
                    DataProtection = UploadLinkObject.Envelope.ExchangeMetadata.DataProtection,
                    ProtectionScheme = UploadLinkObject.Envelope.ExchangeMetadata.ProtectionScheme,
                    DigitalSignatureReference = UploadLinkObject.Envelope.ExchangeMetadata.DigitalSignatureReference,
                    DigitalSignature = new DigitalSignature
                    {
                        DigitalSignatureValue = UploadLinkObject.Envelope.ExchangeMetadata.DigitalSignatureValue.DigitalSignature,
                        PublicCertificate = UploadLinkObject.Envelope.ExchangeMetadata.DigitalSignatureValue.PublicCertificate,
                        PublicRootCertificateThumbprint = UploadLinkObject.Envelope.ExchangeMetadata.DigitalSignatureValue.PublicRootCertificateThumbprint
                    },
                },
                FromSubscription = UploadLinkObject.Envelope.FromSubscription,
                TransactionIdentifier = UploadLinkObject.Envelope.TransactionIdentifier,
                ReceivedTime = DateTime.UtcNow,
                EnvelopeSignatureCertificate = UploadLinkObject.Envelope.EnvelopeSignatureCertificate,
                EnvelopeSignature = UploadLinkObject.EnvelopeSignature,
                EnvelopeRootCertificateThumbprint = UploadLinkObject.Envelope.EnvelopeRootCertificateThumbprint,
                EnvelopeSignatureTime = UploadLinkObject.Envelope.EnvelopeSignatureTime
            };


            var sqlConnectionString = _configuration.GetConnectionString("DataAccessPostgreSqlProvider");

            var ob = new DbContextOptionsBuilder();
            ob.UseNpgsql(
                        sqlConnectionString,
                        b => b.MigrationsAssembly("SECOM.DataAccess")
                    );


            using (var dbContext = new SecomDbContext(ob.Options))
            {
                await dbContext.AddAsync(uploadedMessage);
                await dbContext.SaveChangesAsync();
            }
        }
    }
}
