﻿using System;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using SECOM.Common.Models.Public;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/access")]
    [ApiController]
    public class AccessController : ControllerBase
    {
        /// <summary>
        /// Access to information can be requested through the Access interface.
        /// </summary>
        /// <param name="accessRequestObject">Description of reason for requesting access to information</param>
        /// <returns>Confirmation or error message. Result from request is sent asynchronous through Access Notification interface.</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpPost]
        //[Route("request")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(AccessRequestObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<AccessResponseObject>> RequestAccess([FromBody] AccessRequestObject accessRequestObject)
        {
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new ResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                return Ok(new AccessResponseObject
                {
                    Message = $"The operation was performed successfully"
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// Result from Access Request shall be sent asynchronous through this interface.
        /// </summary>
        /// <param name="accessNotificationRequestObject">Result from the request for access; True or False</param>
        /// <returns>Confirmation or error message</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpPost]
        [Route("notification")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<ResponseObject>> Notification([FromBody] AccessNotificationObject accessNotificationRequestObject)
        {
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new ResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                return Ok(new ResponseObject
                {
                    Message = $"The operation was performed successfully"
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }
    }
}
