﻿using System;
using System.Linq;
using System.Text;
using System.Text.Json;
using System.Threading.Tasks;
using AutoMapper;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using SECOM.Common.Models.Public;
using SECOM.Common.Services;
using SECOM.DataAccess;
using SECOM.DataAccess.Entities;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/acknowledgement")]
    [ApiController]
    public class AcknowledgementController : ControllerBase
    {
        private readonly ILogger _logger;
        private SecomDbContext _dbContext;
        private ISignatureService _signatureService;
        private readonly ICertificateService _certificateService;
        private readonly IMapper _mapper;
        private readonly IHttpContextAccessor _contextAccessor;


        public AcknowledgementController(ILogger<AcknowledgementController> logger, 
            SecomDbContext secomDbContext, ISignatureService signatureService, 
            IMapper mapper, IHttpContextAccessor contextAccessor,
            ICertificateService certificate)
        {
            _logger = logger;
            _signatureService = signatureService;
            _dbContext = secomDbContext;
            _mapper = mapper;
            _contextAccessor = contextAccessor;
            _certificateService = certificate;
        }

        /// <summary>
        /// Upload acknowledgement for previously sent message
        /// </summary>
        /// <param name="acknowledgementObject">Object with reference to information and time when delivered</param>
        /// <returns>Confirmation of upload or error message, including description of incorrect or missing values in payload for the specific service.</returns>
        /// <response code="200">Successfully received ACK for message *transactionIdentifier*</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>         
        /// <response code="403">Not authorized to upload ACK</response>
        /// <response code="405">Method not allowed</response>        
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpPost]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(AcknowledgementResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(AcknowledgementResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<AcknowledgementResponseObject>> Post([FromBody] AcknowledgementObject acknowledgementObject)
        {
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new AcknowledgementResponseObject
                    {
                        Code = Secom_ResponseCodeEnum.MissingRequiredDataForTheService,
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                var x509Cert = _certificateService.GetX509CertFromBase64(acknowledgementObject.Envelope.EnvelopeCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, null, out var errorMessage);

                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new AcknowledgementResponseObject
                    {
                        Code = Secom_ResponseCodeEnum.InvalidCertificate,
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                //Verify signature is optional
                if (acknowledgementObject.DigitalSignature != null)
                {
                    var serializeOptions = new JsonSerializerOptions
                    {
                        PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
                    };

                    var status = _signatureService.VerifyEnvelopeSignature(acknowledgementObject.Envelope, acknowledgementObject.DigitalSignature, 
                        acknowledgementObject.Envelope.EnvelopeCertificate, out string signatureVerificationResult);
                    if (!status)
                    {
                        return BadRequest(new AcknowledgementResponseObject
                        {
                            Code = Secom_ResponseCodeEnum.FailedSignatureVerification,
                            Message = $"{string.Join(", ", signatureVerificationResult)}"
                        });
                    }
                }

                //Store received acknowledgement
                Acknowledgement newAck = ConvertToEntity(acknowledgementObject);
                await _dbContext.AddAsync(newAck);

                // Get caller url
                var sslCert = _contextAccessor.HttpContext.Connection.ClientCertificate;
                _logger.LogDebug($"{sslCert.FriendlyName}, {sslCert.Thumbprint}, {sslCert.Subject}");
                var callerMrn = _certificateService.GetInstanceIdFromCert(sslCert);
                _logger.LogDebug(callerMrn);

                //Create notification
                var notification = new Notification();
                
                notification.Subject = $"New '{acknowledgementObject.Envelope.AckType}' acknowledgement received.";
                notification.Body = $"Transaction: {acknowledgementObject.Envelope.TransactionIdentifier}. {Environment.NewLine}";
                notification.Body += $"Sent from service: {callerMrn}. {Environment.NewLine}";
                notification.Body += $"Received at: {DateTime.Now}. {Environment.NewLine}";
                if(acknowledgementObject.Envelope.NackType.HasValue)
                {
                    notification.Body += $"Error reason: {acknowledgementObject.Envelope.NackType.Value}.";
                }

                notification.ReceivedAt = DateTime.Now;
                notification.TransactionId = acknowledgementObject.Envelope.TransactionIdentifier;

                await _dbContext.AddAsync(notification);
                await _dbContext.SaveChangesAsync();

                return Ok(new AcknowledgementResponseObject
                {
                    Message = $"Successfully received ACK for {acknowledgementObject.Envelope.TransactionIdentifier}."
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new AcknowledgementResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        private Acknowledgement ConvertToEntity(AcknowledgementObject acknowledgementObject)
        {
            return new Acknowledgement
            {
                AckType = (int)acknowledgementObject.Envelope.AckType,
                NackType = (int?)acknowledgementObject.Envelope.NackType,
                TransactionIdentifier = acknowledgementObject.Envelope.TransactionIdentifier,
                CreatedAt = DateTime.UtcNow,
                EnvelopeAckCertificate = acknowledgementObject.Envelope.EnvelopeCertificate,
                EnvelopeRootCertificateThumbprint = acknowledgementObject.Envelope.EnvelopeRootCertificateThumbprint,
                DigitalSignature = acknowledgementObject.DigitalSignature,
                EnvelopeSignatureTime = acknowledgementObject.Envelope.EnvelopeSignatureTime
            };

            
        }
    }
}