﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using SECOM.Common.Models.Public;
using SECOM.DataAccess;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/capability")]
    [ApiController]
    public class CapabilityController : ControllerBase
    {
        private readonly SecomDbContext _dbContext;
        private readonly ILogger _logger;

        public CapabilityController(ILogger<CapabilityController> logger, SecomDbContext context)
        {
            _logger = logger;
            _dbContext = context;
        }

        /// <summary>
        /// The purpose of the interface is to provide a dynamic method to ask a service instance at runtime what interfaces that are accessible, and what payload formats and version that are valid.
        /// </summary>
        /// <returns>Description of the service interface capability. The service interface capability includes flag that indicate which interfaces the service are capable to handle and which information products the service are capable to handle. </returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        [HttpGet]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(CapabilityResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        public async Task<ActionResult<CapabilityResponseObject>> Get()
        {
            try
            {
                var capabilities = new List<CapabilityObject>();
                var caps = _dbContext.Capability.ToList();
                foreach(var cap in caps)
                {
                    var obj = new CapabilityObject
                    {
                        ImplementedInterfaces = new ImplementedInterfaces
                        {
                            Upload = cap.ImplementsUpload,
                            UploadLink = cap.ImplementsUploadLink,
                            Access = cap.ImplementsAccess,
                            EncryptionKey = cap.ImplementsEncryptionKey,
                            Get = cap.ImplementsGet,
                            GetByLink = cap.ImplementsGetByLink,
                            GetSummary = cap.ImplementsGetSummary,
                            Subscription = cap.ImplementsSubscription
                        },
                        DataProductType = (Secom_DataProductTypeEnum)cap.DataProductType,
                        ProductSchemaUrl = cap.ProductSchemaUrl,
                        ServiceVersion = cap.ServiceVersion
                    };
                    capabilities.Add(obj);
                }

                return Ok(new CapabilityResponseObject
                {
                    Capability = capabilities
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }
    }
}
