﻿using System;
using System.Linq;
using System.Threading.Tasks;
using System.Web;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Org.BouncyCastle.Math.EC.Rfc7748;
using SECOM.Common.Models.Public;
using SECOM.Common.Services;
using SECOM.DataAccess;
using SECOM.DataAccess.Entities;
using SECOM.Service.Public.BackgroundTask;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/encryptionKey")]
    [ApiController]
    public class EncryptionKeyController : ControllerBase
    {
        private readonly ILogger _logger;
        private readonly IHttpContextAccessor _contextAccessor;
        private readonly ISignatureService _signatureService;
        private readonly SecomDbContext _dbContext;
        private readonly ICertificateService _certificateService;
        private readonly IBackgroundTaskQueue _taskQueue;
        private readonly IServiceProvider _serviceProvider;

        public EncryptionKeyController(ILogger<EncryptionKeyController> logger,
            IHttpContextAccessor contextAccessor,
            ISignatureService signatureService,
            SecomDbContext dbContext,
            ICertificateService certificateService,
            IBackgroundTaskQueue taskQueue,
            IServiceProvider serviceProvider)
        {
            _logger = logger;
            _contextAccessor = contextAccessor;
            _signatureService = signatureService;
            _dbContext = dbContext;
            _certificateService = certificateService;
            _taskQueue = taskQueue;
            _serviceProvider = serviceProvider;

            _logger.LogDebug($"{GetType().FullName} accessed by user {contextAccessor.HttpContext.User.Identity.Name} from {contextAccessor.HttpContext.Connection.RemoteIpAddress.MapToIPv4()}");
        }

        /// <summary>
        /// This operation requests an encrypted secret key from a producer. The response is sent asynchronously through the consumer’s POST key operation.
        /// </summary>
        /// <param name="encryptionKeyRequestObject">EncryptionKey notification request object</param>
        /// <returns>Response from service</returns>
        /// <response code="202">Request accepted</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Not authorized</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Method not implemented</response>
        [HttpPost]
        [Route("request")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status202Accepted, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(EncryptionKeyResponseObject))]
        public async Task<ActionResult<EncryptionKeyResponseObject>> Get([FromBody] EncryptionKeyRequestObject encryptionKeyRequestObject )
        {
            try
            {
                
                var publicCertificate = encryptionKeyRequestObject.Envelope.PublicCertificate;

                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new EncryptionKeyResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}",
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService
                    });
                }

                // Check envelope certificate
                var x509Cert = _certificateService.GetX509CertFromBase64(encryptionKeyRequestObject.Envelope.EnvelopeSignatureCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, null, out var errorMessage);
                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new EncryptionKeyResponseObject
                    {
                        Message = $"Failed certificate validation: {string.Join(", ", errorMessage)}",
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate
                    });
                }

                // Check envelope signature
                var envelopeSignatureResult = _signatureService.VerifyEnvelopeSignature(encryptionKeyRequestObject.Envelope, encryptionKeyRequestObject.EnvelopeSignature,
                    encryptionKeyRequestObject.Envelope.EnvelopeSignatureCertificate, out var message);
                if (!envelopeSignatureResult)
                {
                    string msg = $"Failed to verify envelope signature for transaction id {encryptionKeyRequestObject.Envelope.DataReference}";
                    _logger.LogWarning(msg);
                    return BadRequest(new EncryptionKeyResponseObject
                    {
                        Message = msg,
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.FailedSignatureVerification
                    });
                }

                //Forward to end application, include parameters + endpoint
                var workItem = _serviceProvider.GetRequiredService<EncryptionKeyWorker>();
                workItem.PublicCertificate = _certificateService.GetX509CertFromBase64(publicCertificate);
                workItem.ClientCertificate = _contextAccessor.HttpContext.Connection.ClientCertificate;
                workItem.DataReference = encryptionKeyRequestObject.Envelope.DataReference;

                var dbEntity = await _dbContext.TemporaryKey
                    .FirstOrDefaultAsync(x => x.DataReference == encryptionKeyRequestObject.Envelope.DataReference);

                workItem.StoredKey = dbEntity;

                if (_contextAccessor.HttpContext.Request.Host.Host.ToLower().Contains("localhost"))
                {
                    var urls = Environment.GetEnvironmentVariable("ASPNETCORE_URLS")?.Split(";");
                    workItem.Url = $"{urls[0]}/v1/EncryptionKey";
                }
                else
                {
                    workItem.Url = string.Empty;
                }

                _taskQueue.QueueBackgroundWorkItem(workItem.SendKey);

                return StatusCode(StatusCodes.Status202Accepted, new EncryptionKeyResponseObject
                {
                    Message = "Request accepted"
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new EncryptionKeyResponseObject
                {
                    Message = $"{ex.Message}",
                    SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService
                });
            }
        }

        /// <summary>
        /// The purpose of the interface is to exchange a temporary secret key.
        /// </summary>
        /// <param name="encryptionKeyObject">Object with symmetric encryption key</param>
        /// <returns>Response from service</returns>
        /// <response code="200">EncryptionKeyResponseObject</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Not authorized to upload</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Method not implemented</response>
        [HttpPost]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(EncryptionKeyResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(EncryptionKeyResponseObject))]
        public async Task<ActionResult<EncryptionKeyResponseObject>> Post([FromBody] EncryptionKeyObject encryptionKeyObject)
        {
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new EncryptionKeyResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}",
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService
                    });
                }

                // Check envelope certificate
                var x509Cert = _certificateService.GetX509CertFromBase64(encryptionKeyObject.Envelope.EnvelopeSignatureCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, null, out var errorMessage);
                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new EncryptionKeyResponseObject
                    {
                        Message = $"Failed certificate validation: {string.Join(", ", errorMessage)}",
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate
                    });
                }

                // Check envelope signature
                var envelopeSignatureResult = _signatureService.VerifyEnvelopeSignature(encryptionKeyObject.Envelope, encryptionKeyObject.EnvelopeSignature, 
                    encryptionKeyObject.Envelope.EnvelopeSignatureCertificate, out var message);
                if(!envelopeSignatureResult)
                {
                    string msg = $"Failed to verify envelope signature for transaction id {encryptionKeyObject.Envelope.TransactionIdentifier}";
                    _logger.LogWarning(msg);
                    return BadRequest(new EncryptionKeyResponseObject
                    {
                        Message = msg,
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.FailedSignatureVerification
                    });
                }
                var key = _dbContext.EncryptionKey.FirstOrDefault(x => x.TransactionIdentifier == encryptionKeyObject.Envelope.TransactionIdentifier);
                if (key == null)
                {
                    key = new EncryptionKey
                    {
                        DigitalSignature = new DigitalSignature
                        {
                            DigitalSignatureValue = encryptionKeyObject.Envelope.DigitalSignatureValue.DigitalSignature,
                            PublicCertificate = encryptionKeyObject.Envelope.DigitalSignatureValue.PublicCertificate,
                            PublicRootCertificateThumbprint = encryptionKeyObject.Envelope.DigitalSignatureValue.PublicRootCertificateThumbprint
                        },
                        IV = encryptionKeyObject.Envelope.IV,
                        Key = encryptionKeyObject.Envelope.EncryptionKey,
                        TransactionIdentifier = encryptionKeyObject.Envelope.TransactionIdentifier,
                        EnvelopeRootCertificateThumbprint = encryptionKeyObject.Envelope.EnvelopeRootCertificateThumbprint,
                        EnvelopeSignature = encryptionKeyObject.EnvelopeSignature,
                        EnvelopeSignatureCertificate = encryptionKeyObject.Envelope.EnvelopeSignatureCertificate,
                        EnvelopeSignatureTime = encryptionKeyObject.Envelope.EnvelopeSignatureTime,
                    };

                    _dbContext.Add(key);
                    _dbContext.SaveChanges();
                }
                else
                {
                    string msg = $"A key has already been uploaded for transaction id {encryptionKeyObject.Envelope.TransactionIdentifier}";
                    _logger.LogError(msg);
                    return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                    {
                        Message = msg
                    });
                }

               
                return Ok(new EncryptionKeyResponseObject
                {
                    Message = $"Encryption key uploaded successfully."
                });
            }
            catch (Exception ex)
            {
                _logger.LogError($"{ex}");
                return StatusCode(StatusCodes.Status500InternalServerError, new EncryptionKeyResponseObject
                {
                    Message = $"{ex.Message}",
                    SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService
                });
            }
        }
    }
}
