﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Logging;
using SECOM.Common.Models.Public;
using SECOM.Common.Services;
using SECOM.DataAccess;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/object")]
    [ApiController]
    public class GetController : ControllerBase
    {
        private SecomDbContext _dbContext;
        private readonly ILogger _logger;
        private readonly ICertificateService _certificateService;
        private const int _maxPageSize = 100;

        public GetController(SecomDbContext context, 
            ILogger<GetController> logger,
            ICertificateService certificateService)
        {
            _dbContext = context;
            _logger = logger;
            _certificateService = certificateService;
        }

        /// <summary>
        /// The Get interface is used for pulling information from a service provider. 
        /// </summary>
        /// <param name="dataReference ">Reference to information object, e.g. from Get Summary</param>
        /// <param name="containerType">Data Type requested </param>
        /// <param name="dataProductType">Data product type name</param>
        /// <param name="productVersion">S-100 based Product specification version</param>
        /// <param name="geometry">Geometry condition for geo-located information objects as WKT LineString or Polygon</param>
        /// <param name="unlocode">See UN web page</param>
        /// <param name="validFrom">Time related to validity period start for information object</param>
        /// <param name="validTo">Time related to validity period end for information object</param>
        /// <param name="page">Requested pagination page</param>
        /// <param name="pageSize">Requested pagination page size</param>
        /// <returns>Set of messages package with metadata or an error message</returns>
        /// <response code="200">GetResponseObject</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="403">Not authorized to requested information</response>
        /// <response code="404">Information not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpGet]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(GetResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<GetResponseObject>> Get(
            Guid? dataReference,
            ContainerTypeEnum? containerType,
            Secom_DataProductTypeEnum? dataProductType,
            string productVersion,
            string geometry,
            string unlocode,
            DateTime? validFrom = null,
            DateTime? validTo = null,
            int page = 1,
            int pageSize = _maxPageSize)
        {
            try
            { 
                //TODO: Add validation of parameters i.e. if not valid return bad request

                var result = new GetResponseObject();
                //IEnumerable<DataAccess.Entities.PublishedMessage> query = _dbContext.PublishedMessage;
                IQueryable<DataAccess.Entities.PublishedMessage> query = _dbContext.PublishedMessage.Include(ex => ex.ExchangeMetadata).Include(d => d.ExchangeMetadata.DigitalSignature);
                var responseMessage = string.Empty;
                
                
                if(dataReference.HasValue)
                {
                    query = query.Where(p => p.DataReference == dataReference.Value);
                    responseMessage += $", dataReference = {dataReference.Value}";
                }

                if (containerType.HasValue)
                {
                    var containerTypeValue = (int)containerType.Value;
                    query = query.Where(p => p.ContainerType == containerTypeValue);
                    responseMessage += $", containerType = {containerTypeValue}";
                }

                if (dataProductType.HasValue)
                {
                    query = query.Where(x => x.DataProductTypeId == (int)dataProductType.Value);
                    responseMessage += $", dataProductType = {dataProductType.Value}";
                }

                if (!string.IsNullOrEmpty(productVersion))
                {
                    query = query.Where(x => x.ProductVersion == productVersion);
                    responseMessage += $", productVersion = {productVersion}";
                }

                //if(!string.IsNullOrEmpty(geometry))
                //{
                //    //Fix me! Need to add logic for geo search
                //    query = query.Where(x => x.Geometry == geometry);
                //}

                if (!string.IsNullOrEmpty(unlocode))
                {
                    query = query.Where(x => x.Unlocode == unlocode);
                    responseMessage += $", unlocode = {unlocode}";
                }

                if(validFrom.HasValue)
                {
                    query = query.Where(x => x.CreatedTime > validFrom.Value);
                    responseMessage += $", validFrom = {validFrom.Value}";
                }

                if (validTo.HasValue)
                {
                    query = query.Where(x => x.CreatedTime < validTo.Value);
                    responseMessage += $", validTo = {validTo.Value}";
                }

                // Get identity ether from internal id talbe or from id registry
                SECOM.DataAccess.Entities.Identity identity = GetCallerIdentity();

                var totItems = query.Count();
                if (page == 0) ////If the page parameter is set to 0, the server shall not return any items, but shall return maxItemsPerPage.
                {
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };
                    return Ok(result);
                }
                else
                {
                    var responseObjects = new List<DataResponseObject>();
                    responseObjects = await GetPublishedMessages(identity, query, page, Math.Min(_maxPageSize, pageSize));
                    if (responseObjects == null || responseObjects.Count == 0)
                    {
                        string message = "";
                        if (string.IsNullOrEmpty(responseMessage)) message = $"Information not found.";
                        else if (dataReference.HasValue) message = $"Information with reference {dataReference.Value} not found.";
                        else message = $"Information for query {responseMessage} not found.";

                        return StatusCode(StatusCodes.Status404NotFound, new ResponseObject
                        {
                            Message = message
                        });
                    }

                    //1. Check if data should be encrypted

                    //2. Get service instance id

                    //3. Get public key to origin requester, how? Extend input parameters and include publicKey as Base64

                    //4. FindService(instanceId)

                    //5. Encrypt data

                    //6. Add encrypted data to response

                    //7. Start new thread and consume EncryptionKey with encryption key object

                    result.DataResponseObject = responseObjects;
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };

                    return Ok(result);
                }
 
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// A list of information shall be returned from this interface. The list contains identity, status and short description of each information object. 
        /// </summary>
        /// <param name="containerType">Data Type requested </param>
        /// <param name="dataProductType">Data product type name</param>
        /// <param name="productVersion">S-100 based Product specification version</param>
        /// <param name="geometry">Geometry condition for geo-located information objects</param>
        /// <param name="unlocode">See UN web page</param>
        /// <param name="validFrom">Time related to validity period start for information object</param>
        /// <param name="validTo">Time related to validity period end for information object</param>
        /// <param name="page">Requested pagination page</param>
        /// <param name="pageSize">Requested pagination page size</param>
        /// <returns>Set of messages package with metadata or an error message</returns>
        /// <response code="200">GetSummaryResponseObject</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="403">Not authorized to requested information</response>
        /// <response code="404">Information not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpGet]
        [Route("summary")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(GetSummaryResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<GetSummaryResponseObject>> GetSummary(
            ContainerTypeEnum? containerType,
            Secom_DataProductTypeEnum? dataProductType,
            string productVersion,
            string geometry,
            string unlocode,
            DateTime? validFrom = null,
            DateTime? validTo = null,
            int page = 1,
            int pageSize = _maxPageSize)
        {
            try
            {

                IQueryable<DataAccess.Entities.PublishedMessage> query = _dbContext.PublishedMessage.Include(ex => ex.ExchangeMetadata).Include(d => d.ExchangeMetadata.DigitalSignature);
                var result = new GetSummaryResponseObject();
                var responseMessage = string.Empty;

                //TODO: Add data type based on implementation
                if(containerType.HasValue)
                {
                    var containerTypeValue = (int)containerType.Value;
                    query = query.Where(p => p.ContainerType == containerTypeValue);
                    responseMessage += $", containerType = {containerType.Value}";
                }

                if (dataProductType.HasValue)
                {
                    var val = (int)dataProductType.Value;
                    query = query.Where(x => x.DataProductTypeId == val);
                    responseMessage += $", productSpecificationId = {dataProductType.Value}";
                }

                if (!string.IsNullOrEmpty(productVersion))
                {
                    query = query.Where(x => x.ProductVersion == productVersion);
                    responseMessage += $", productSpecificationVersion = {productVersion}";
                }

                if (!string.IsNullOrEmpty(unlocode))
                {
                    query = query.Where(x => x.Unlocode == unlocode);
                    responseMessage += $", unlocode = {unlocode}";
                }

                if (validFrom.HasValue)
                {
                    var from = validFrom.Value;
                    query = query.Where(x => x.CreatedTime > from);
                    responseMessage += $", validFrom = {from}";
                }

                if (validTo.HasValue)
                {
                    var to = validTo.Value;
                    query = query.Where(x => x.CreatedTime < to);
                    responseMessage += $", validTo = {to}";
                }

                var totItems = query.Count();
                if (page == 0) ////If the page parameter is set to 0, the server shall not return any items, but shall return maxItemsPerPage.
                {
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };
                    return Ok(result);
                }

                else
                {
                    var summaryObjects = await GetPublishedMessagesSummary(query, page, Math.Min(_maxPageSize, pageSize));

                    if (summaryObjects == null || summaryObjects.Count == 0)
                    {
                        return StatusCode(StatusCodes.Status404NotFound, new ResponseObject
                        {
                            Message = string.IsNullOrEmpty(responseMessage) ? "Information not found." : $"Information for query {responseMessage} not found."
                        });
                    }

                    result.InformationSummaryObject = summaryObjects;
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };

                    return Ok(result);
                }
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// The Get By Link interface is used for pulling information from a data storage handled by the information owner. 
        /// </summary>
        /// <param name="transactionIdentifier">Indentifier from the uploaded link</param>
        /// <returns>Confirmation of upload or error message, including description of incorrect or missing values in payload for the specific service.</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="403">Not authorized to requested information</response>
        /// <response code="404">Information with reference *dataReference* not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpGet]
        [Route("link")]
        [Consumes("application/json")]
        [Produces("application/octet-stream", new[] { "application/json"})]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(byte[]))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<Stream>> GetLink(Guid transactionIdentifier)
        {
            try
            {
                var x509Cert = HttpContext.Connection.ClientCertificate;
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, null, out var errorMessage);
                if (!isValidCert)
                {
                    _logger.LogError($"Invalid certificate {x509Cert.FriendlyName}: { string.Join(", ", errorMessage)}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate,
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                var callerMrn = _certificateService.GetInstanceIdFromCert(x509Cert);
                var obj = _dbContext.SentMessage.FirstOrDefault(x =>
                    x.TransactionIdentifier == transactionIdentifier
                    && x.TimeToLive > DateTime.Now);

                if (obj == null)
                    return NotFound(new ResponseObject { 
                        Message = $"Information with reference {transactionIdentifier} not found"
                    });

                if (obj.ReceiverMrn == null || obj.ReceiverMrn.ToLower() != callerMrn.ToLower())
                    return Forbid();

                return Ok(new MemoryStream(obj.Data));
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        #region private
        private async Task<List<DataResponseObject>> GetPublishedMessages(DataAccess.Entities.Identity identity, IQueryable<DataAccess.Entities.PublishedMessage> query, int page, int pageSize)
        {
            var result = new List<DataResponseObject>();

            var data = await query.OrderBy(x => x.CreatedTime).Skip((page - 1) * pageSize).Take(pageSize).ToListAsync();
            foreach (var row in data)
            {
                // Now look up if orgId is authorized to this voyageplan
                //var identity = _identityService.GetCallerIdentity();
                //var aclObject = _dbContext.AclObject.Where(x => x.Subscriber.Id == identity.Id && x.MessageId == row.DataReference);

                var obj = new DataResponseObject
                {
                    Data = row.Data,
                    AckRequest = (AckRequestEnum)row.AckRequest,
                    ExchangeMetadata = new ExchangeMetadata
                    {
                        CompressionFlag = row.ExchangeMetadata.CompressionFlag,
                        DataProtection = row.ExchangeMetadata.DataProtection,
                        DigitalSignatureReference = row.ExchangeMetadata.DigitalSignatureReference,
                        ProtectionScheme = row.ExchangeMetadata.ProtectionScheme,
                        DigitalSignatureValue = new DigitalSignatureValueObject
                        {
                            DigitalSignature = row.ExchangeMetadata.DigitalSignature.DigitalSignatureValue,
                            PublicCertificate = row.ExchangeMetadata.DigitalSignature.PublicCertificate,
                            PublicRootCertificateThumbprint = row.ExchangeMetadata.DigitalSignature.PublicRootCertificateThumbprint
                        }
                    }
                };

                result.Add(obj);
            }
            return result;
        }

        private async Task<List<SummaryObject>> GetPublishedMessagesSummary(IQueryable<DataAccess.Entities.PublishedMessage> query, int page, int pageSize)
        {
            var result = new List<SummaryObject>();

            var data = await query.OrderBy(x => x.CreatedTime).Skip((page - 1) * pageSize).Take(pageSize).ToListAsync();

            foreach (var row in data)
            {
                var obj = new SummaryObject
                {
                    DataReference = row.DataReference.ToString(),
                    DataProtection = row.ExchangeMetadata.DataProtection,
                    DataCompression = row.ExchangeMetadata.CompressionFlag,
                    ContainerType = (ContainerTypeEnum)Enum.Parse(typeof(ContainerTypeEnum), _dbContext.ContainerTypeEnum.Where(x => x.Id == row.ContainerType).FirstOrDefault().Value, true),
                    DataProductType = (Secom_DataProductTypeEnum)row.DataProductTypeId,
                    Info_productVersion = row.ProductVersion,
                    Info_size = row.Data.Length,
                };
                result.Add(obj);
            }

            return result;
        }

        private DataAccess.Entities.Identity GetCallerIdentity()
        {
            var result = new DataAccess.Entities.Identity();

            X509Certificate2 cert = Request.HttpContext.Connection.ClientCertificate;
            if (cert != null)
            {
                var subj = cert.Subject;
                var thumbp = cert.Thumbprint;
                var serial = cert.SerialNumber;

            }
            else
            {
                //return Content("....", "application/json");
            }

            return result;
        }

        #endregion
    }
}
