﻿using System;
using System.Linq;
using System.Text;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Microsoft.EntityFrameworkCore;
using SECOM.Common.Models.Public;
using SECOM.Common.Schema;
using SECOM.Common.Services;
using SECOM.DataAccess;
using SECOM.DataAccess.Entities;
using SECOM.Service.Public.BackgroundTask;
using System.Security.Cryptography.X509Certificates;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/rtz")]
    [ApiController]
    public class RtzController : ControllerBase
    {
        private readonly ILogger _logger;
        private readonly IHttpContextAccessor _contextAccessor;
        private readonly ISignatureService _signatureVerificationService;
        private readonly IIdentityRegistryService _identityRegistryService;
        private readonly IServiceRegistryService _serviceRegistryService;
        private readonly ICertificateService _certificateService;
        private readonly IBackgroundTaskQueue _taskQueue;
        private readonly IServiceProvider _serviceProvider;
        private readonly SecomDbContext _dbContext;
        private const int _maxPageSize = 100;

        public RtzController(ILogger<RtzController> logger,
            IHttpContextAccessor contextAccessor,
            ISignatureService signatureVerificationService,
            IIdentityRegistryService identityRegistryService,
            IServiceRegistryService serviceRegistryService,
            ICertificateService certificateService,
            IBackgroundTaskQueue taskQueue,
            IServiceProvider serviceProvider,
            SecomDbContext dbContext)
        {
            _logger = logger;
            _contextAccessor = contextAccessor;
            _signatureVerificationService = signatureVerificationService;
            _dbContext = dbContext;
            _identityRegistryService = identityRegistryService;
            _serviceRegistryService = serviceRegistryService;
            _certificateService = certificateService;
            _taskQueue = taskQueue;
            _serviceProvider = serviceProvider;

            _logger.LogDebug($"{GetType().FullName} accessed by user {contextAccessor.HttpContext.User.Identity.Name} from {contextAccessor.HttpContext.Connection.RemoteIpAddress}");
        }

        /// <summary>
        /// The interface shall be used for uploading (pushing) data to a consumer. The operation expects one single data object and its metadata.
        /// </summary>
        /// <param name="uploadObject">The data (payload) package with its metadata</param>
        /// <returns>Confirmation of upload or error message, including description of incorrect or missing values in payload for the specific service.</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Forbidden</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="413">Request entity too large</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>        
        [HttpPost]
        [Consumes("application/json")]
        [Produces("application/json")]
        [RequestSizeLimit(350000)]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status413RequestEntityTooLarge, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(UploadResponseObject))]

        public async Task<ActionResult<UploadResponseObject>> Post([FromBody] UploadObject uploadObject)
        {
            var response = string.Empty;

            try
            {

                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService,
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                // Validate data type, only possible if not encrypted
                //if (!uploadObject.Envelope.ExchangeMetadata.DataProtection)
                //{
                //    try
                //    {
                //        var validator = new SchemaValidationService();
                //        //TODO: Parse message from inside a dataset
                //        var data = Encoding.UTF8.GetString(uploadObject.Envelope.Data);

                //        validator.ValidateRtzXml(data);
                //    }
                //    catch (Secom_SchemaValidationException ex)
                //    {
                //        return BadRequest(new UploadResponseObject
                //        {
                //            Message = $"{ex.Message}",
                //            SECOM_ResponseCode = Secom_ResponseCodeEnum.XmlSchemaValidationError
                //        });
                //    }
                //}

                //Need to validate client certificate from TLS

                //Verify claimed public certificate in envelope against root thumbprint
                //TODO How do we verify the remaining certs, TLS, data 
                //Need thumbprint for verify TLS cert against e.g. Navelink
                var x509Cert = _certificateService.GetX509CertFromBase64(uploadObject.Envelope.EnvelopeSignatureCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, uploadObject.Envelope.EnvelopeRootCertificateThumbprint, out var errorMessage);
                response += errorMessage;
                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate,
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                //Check envelope signature
                var isValid = _signatureVerificationService.VerifyEnvelopeSignature(uploadObject.Envelope, uploadObject.EnvelopeSignature,
                    uploadObject.Envelope.EnvelopeSignatureCertificate, out var message);
                response += message;

                if (!isValid)
                {
                    _logger.LogError($"Failed to verify envelope signature for message id {uploadObject.Envelope.TransactionIdentifier}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.FailedSignatureVerification,
                        Message = $"{string.Join(", ", message)}"
                    });
                }

                var uploadedMessage = new UploadedMessage
                {
                    AckEndpoint = await GetCallerEndpoint(),
                    AckRequest = (int)uploadObject.Envelope.AckRequest,
                    Data = uploadObject.Envelope.Data,
                    DataType = (int)Common.Models.Public.DataTypeEnum.RTZ,
                    ExchangeMetadata = new SECOM.DataAccess.Entities.ExchangeMetadata
                    {
                        CompressionFlag = uploadObject.Envelope.ExchangeMetadata.CompressionFlag,
                        DataProtection = uploadObject.Envelope.ExchangeMetadata.DataProtection,
                        ProtectionScheme = uploadObject.Envelope.ExchangeMetadata.ProtectionScheme,
                        DigitalSignatureReference = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureReference,
                        DigitalSignature = new DigitalSignature
                        {
                            DigitalSignatureValue = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureValue.DigitalSignature,
                            PublicCertificate = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureValue.PublicCertificate,
                            PublicRootCertificateThumbprint = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureValue.PublicRootCertificateThumbprint
                        },
                    },
                    FromSubscription = uploadObject.Envelope.FromSubscription,
                    TransactionIdentifier = uploadObject.Envelope.TransactionIdentifier,
                    ReceivedTime = DateTime.UtcNow,
                    EnvelopeSignatureCertificate = uploadObject.Envelope.EnvelopeSignatureCertificate,
                    EnvelopeRootCertificateThumbprint = uploadObject.Envelope.EnvelopeRootCertificateThumbprint,
                    EnvelopeSignature = uploadObject.EnvelopeSignature,
                    SignatureValidatedOk = isValid,
                    EnvelopeSignatureTime = uploadObject.Envelope.EnvelopeSignatureTime
                };

                await _dbContext.AddAsync(uploadedMessage);
                await _dbContext.SaveChangesAsync();

                _logger.LogInformation($"Successfully added uploaded message {uploadedMessage.TransactionIdentifier} to database.");
                return Ok(new UploadResponseObject
                {
                    Message = "Message was successfully uploaded"
                });
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, ex.Message);
                if (ex.InnerException != null) _logger.LogError($"{ex.InnerException}");
                if (!string.IsNullOrEmpty(ex.StackTrace)) _logger.LogError(ex.StackTrace);
                return StatusCode(StatusCodes.Status500InternalServerError, new UploadResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }


        /// <summary>
        /// The purpose with this interface is to upload (push) a link to information to a consumer. 
        /// </summary>
        /// <param name="uploadLinkObject">The message link with its metadata</param>
        /// <returns>Confirmation of upload or error message</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Forbidden</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response> 	
        [HttpPost]
        [Route("link")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(UploadResponseObject))]
        public async Task<ActionResult<UploadResponseObject>> PostLink([FromBody] UploadLinkObject uploadLinkObject)
        {
            var response = string.Empty;
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService,
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                // Get caller url
                var sslCert = _contextAccessor.HttpContext.Connection.ClientCertificate;
                _logger.LogDebug($"{sslCert.FriendlyName}, {sslCert.Thumbprint}, {sslCert.Subject}");
                var callerMrn = _certificateService.GetInstanceIdFromCert(sslCert);
                _logger.LogDebug(callerMrn);

                //var serviceInstance = await _serviceRegistryService.FindService($"instanceId:{callerMrn.Replace(":", "\\:")}", "", 0, 1000);

                //Verify claimed public certificate in envelope against root thumbprint
                //TODO How do we verify the remaining certs, TLS, data 
                //Need thumbprint for verify TLS cert against e.g. Navelink
                var x509Cert = _certificateService.GetX509CertFromBase64(uploadLinkObject.Envelope.EnvelopeSignatureCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, uploadLinkObject.Envelope.EnvelopeRootCertificateThumbprint, out var errorMessage);

                response += errorMessage;
                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate,
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                // Check envelope signature
                var isValid = _signatureVerificationService.VerifyEnvelopeSignature(uploadLinkObject.Envelope, uploadLinkObject.EnvelopeSignature,
                    uploadLinkObject.Envelope.EnvelopeSignatureCertificate, out var message);
                response += message;

                if (!isValid)
                {
                    _logger.LogError($"Failed to verify envelope signature for message id {uploadLinkObject.Envelope.TransactionIdentifier}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.FailedSignatureVerification,
                        Message = $"{string.Join(", ", message)}"
                    });
                }

                var workItem = _serviceProvider.GetRequiredService<UploadLinkWorker>();

                workItem.ClientCertificate = _contextAccessor.HttpContext.Connection.ClientCertificate;
                workItem.UploadLinkObject = uploadLinkObject;
                workItem.AckEndpoint = await GetCallerEndpoint();
                workItem.DataType = Common.Models.Public.DataTypeEnum.RTZ;

                _taskQueue.QueueBackgroundWorkItem(workItem.GetLink);

                return Ok(new UploadResponseObject
                {
                    Message = $"Successfully received link with transactionId: {uploadLinkObject.Envelope.TransactionIdentifier}"
                });
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, ex.Message);
                return StatusCode(StatusCodes.Status500InternalServerError, new UploadResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// The Get interface is used for pulling information from a service provider. 
        /// </summary>
        /// <param name="dataReference ">Reference to information object, e.g. from Get Summary</param>
        /// <param name="geometry">Geometry condition for geo-located information objects as WKT LineString or Polygon</param>
        /// <param name="unlocode">See UN web page</param>
        /// <param name="validFrom">Time related to validity period start for information object</param>
        /// <param name="validTo">Time related to validity period end for information object</param>
        /// <param name="page">Requested pagination page</param>
        /// <param name="pageSize">Requested pagination page size</param>
        /// <returns>Set of messages package with metadata or an error message</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="403">Not authorized</response>
        /// <response code="404">Information not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpGet]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(GetResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<GetResponseObject>> Get(
            Guid? dataReference,
            string geometry,
            string unlocode,
            DateTime? validFrom = null,
            DateTime? validTo = null,
            int page = 1,
            int pageSize = _maxPageSize)
        {
            try
            {
                //TODO: Add validation of parameters i.e. if not valid return bad request

                var result = new GetResponseObject();
                //IEnumerable<DataAccess.Entities.PublishedMessage> query = _dbContext.PublishedMessage;
                IQueryable<DataAccess.Entities.PublishedMessage> query = _dbContext.PublishedMessage.Include(ex => ex.ExchangeMetadata).Include(d => d.ExchangeMetadata.DigitalSignature);
                var responseMessage = string.Empty;


                if (dataReference.HasValue)
                {
                    query = query.Where(p => p.DataReference == dataReference.Value);
                    responseMessage += $", dataReference = {dataReference.Value}";
                }

                //if (dataType.HasValue)
                //{
                //    var dataTypeValue = (int)dataType.Value;
                //    query = query.Where(p => p.DataType == dataTypeValue);
                //    responseMessage += $", dataType = {dataTypeValue}";
                //}

                if (!string.IsNullOrEmpty(unlocode))
                {
                    query = query.Where(x => x.Unlocode == unlocode);
                    responseMessage += $", unlocode = {unlocode}";
                }

                if (validFrom.HasValue)
                {
                    query = query.Where(x => x.CreatedTime > validFrom.Value);
                    responseMessage += $", validFrom = {validFrom.Value}";
                }

                if (validTo.HasValue)
                {
                    query = query.Where(x => x.CreatedTime < validTo.Value);
                    responseMessage += $", validTo = {validTo.Value}";
                }

                // Get identity ether from internal id talbe or from id registry
                SECOM.DataAccess.Entities.Identity identity = GetCallerIdentity();

                var totItems = query.Count();
                if (page == 0) ////If the page parameter is set to 0, the server shall not return any items, but shall return maxItemsPerPage.
                {
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };
                    return Ok(result);
                }
                else
                {
                    var responseObjects = new List<DataResponseObject>();
                    responseObjects = await GetPublishedMessages(identity, query, page, Math.Min(_maxPageSize, pageSize));
                    if (responseObjects == null || responseObjects.Count == 0)
                    {
                        string message = "";
                        if (string.IsNullOrEmpty(responseMessage)) message = $"Information not found";
                        else if (dataReference.HasValue) message = $"Information with reference {dataReference.Value} not found.";
                        else message = $"Information for query {responseMessage} not found.";

                        return StatusCode(StatusCodes.Status404NotFound, new ResponseObject
                        {
                            Message = message
                        });
                    }

                    //1. Check if data should be encrypted

                    //2. Get service instance id

                    //3. Get public key to origin requester, how? Extend input parameters and include publicKey as Base64

                    //4. FindService(instanceId)

                    //5. Encrypt data

                    //6. Add encrypted data to response

                    //7. Start new thread and consume EncryptionKey with encryption key object

                    result.DataResponseObject = responseObjects;
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };

                    return Ok(result);
                }

            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// The Get By Link interface is used for pulling information from a data storage handled by the information owner. 
        /// </summary>
        /// <param name="TransactionIdentifier">Indentifier from the uploaded link</param>
        /// <returns>Confirmation of upload or error message, including description of incorrect or missing values in payload for the specific service.</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="403">Not authorized</response>
        /// <response code="404">Information not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpGet]
        [Route("link")]
        [Consumes("application/json")]
        [Produces("application/octet-stream", new[] { "application/json" })]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(byte[]))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<Stream>> GetLink(Guid TransactionIdentifier)
        {
            try
            {
                var x509Cert = HttpContext.Connection.ClientCertificate;
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, null, out var errorMessage);
                if (!isValidCert)
                {
                    _logger.LogError($"Invalid certificate {x509Cert.FriendlyName}: { string.Join(", ", errorMessage)}.");
                    return BadRequest(new UploadResponseObject
                    {
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                var callerMrn = _certificateService.GetInstanceIdFromCert(x509Cert);
                var obj = _dbContext.SentMessage.FirstOrDefault(x =>
                    x.TransactionIdentifier == TransactionIdentifier
                    && x.TimeToLive > DateTime.Now);

                if (obj == null)
                    return NotFound();

                if (obj.ReceiverMrn == null || obj.ReceiverMrn.ToLower() != callerMrn.ToLower())
                    return Forbid();

                return Ok(new MemoryStream(obj.Data));
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// A list of information shall be returned from this interface. The list contains identity, status and short description of each information object. 
        /// </summary>
        /// <param name="geometry">Geometry condition for geo-located information objects</param>
        /// <param name="unlocode">See UN web page</param>
        /// <param name="validFrom">Time related to validity period start for information object</param>
        /// <param name="validTo">Time related to validity period end for information object</param>
        /// <param name="page">Requested pagination page</param>
        /// <param name="pageSize">Requested pagination page size</param>
        /// <returns>Set of messages package with metadata or an error message</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="403">Not authorized</response>
        /// <response code="404">Information not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpGet]
        [Route("summary")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(GetSummaryResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<GetSummaryResponseObject>> GetSummary(
            string geometry,
            string unlocode,
            DateTime? validFrom = null,
            DateTime? validTo = null,
            int page = 1,
            int pageSize = _maxPageSize)
        {
            try
            {

                IQueryable<DataAccess.Entities.PublishedMessage> query = _dbContext.PublishedMessage.Include(ex => ex.ExchangeMetadata).Include(d => d.ExchangeMetadata.DigitalSignature);
                var result = new GetSummaryResponseObject();
                var responseMessage = string.Empty;
                var dt = _dbContext.DataTypeEnum.Where(x => x.Value == "RTZ").FirstOrDefault().Id;
                query = query.Where(x => x.DataType == dt);
                if (!string.IsNullOrEmpty(unlocode))
                {
                    query = query.Where(x => x.Unlocode == unlocode);
                    responseMessage += $", unlocode = {unlocode}";
                }

                if (validFrom.HasValue)
                {
                    query = query.Where(x => x.CreatedTime > validFrom.Value);
                    responseMessage += $", validFrom = {validFrom.Value}";
                }

                if (validTo.HasValue)
                {
                    query = query.Where(x => x.CreatedTime < validTo.Value);
                    responseMessage += $", validTo = {validTo.Value}";
                }

                var totItems = query.Count();
                if (page == 0) ////If the page parameter is set to 0, the server shall not return any items, but shall return maxItemsPerPage.
                {
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };
                    return Ok(result);
                }

                else
                {
                    var summaryObjects = await GetPublishedMessagesSummary(query, page, Math.Min(_maxPageSize, pageSize));

                    if (summaryObjects == null || summaryObjects.Count == 0)
                    {
                        return StatusCode(StatusCodes.Status404NotFound, new ResponseObject
                        {
                            Message = string.IsNullOrEmpty(responseMessage) ? "Information not found." : $"Information for query {responseMessage} not found."
                        });
                    }

                    result.InformationSummaryObject = summaryObjects;
                    result.Pagination = new PaginationObject
                    {
                        MaxItemsPerPage = Math.Min(pageSize, _maxPageSize),
                        TotalItems = totItems
                    };

                    return Ok(result);
                }
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }



        #region private

        private async Task<string> GetCallerEndpoint()
        {
            string endPoint = "";
            //For debugging use localhost
            //if (System.Diagnostics.Debugger.IsAttached)
            //{
            //    var urls = Environment.GetEnvironmentVariable("ASPNETCORE_URLS")?.Split(";");
            //    return $"{urls[0]}/v1/Acknowledgement/";
            //}
            // Get caller url
            var callerMrn = _certificateService.GetInstanceIdFromCert(_contextAccessor.HttpContext.Connection.ClientCertificate);
            var serviceInstance = await _serviceRegistryService.FindService($"instanceId:{callerMrn.Replace(":", "\\:")}", "", 0, 1000);

            if (serviceInstance != null && serviceInstance.Count == 1)
                endPoint = $"{serviceInstance.FirstOrDefault().EndpointUri}/v1/Acknowledgement/";

            _logger.LogInformation($"Stored caller acknowledgement endpoint: {endPoint}.");
            return endPoint;
        }

        private DataAccess.Entities.Identity GetCallerIdentity()
        {
            var result = new DataAccess.Entities.Identity();

            X509Certificate2 cert = Request.HttpContext.Connection.ClientCertificate;
            if (cert != null)
            {
                var subj = cert.Subject;
                var thumbp = cert.Thumbprint;
                var serial = cert.SerialNumber;

            }
            else
            {
                //return Content("....", "application/json");
            }

            return result;
        }

        private async Task<List<DataResponseObject>> GetPublishedMessages(DataAccess.Entities.Identity identity, IQueryable<DataAccess.Entities.PublishedMessage> query, int page, int pageSize)
        {
            var result = new List<DataResponseObject>();

            var data = await query.OrderBy(x => x.CreatedTime).Skip((page - 1) * pageSize).Take(pageSize).ToListAsync();
            foreach (var row in data)
            {
                // Now look up if orgId is authorized to this voyageplan
                //var identity = _identityService.GetCallerIdentity();
                //var aclObject = _dbContext.AclObject.Where(x => x.Subscriber.Id == identity.Id && x.MessageId == row.DataReference);

                var obj = new DataResponseObject
                {
                    Data = row.Data,
                    ExchangeMetadata = new Common.Models.Public.ExchangeMetadata
                    {
                        CompressionFlag = row.ExchangeMetadata.CompressionFlag,
                        DataProtection = row.ExchangeMetadata.DataProtection,
                        DigitalSignatureReference = row.ExchangeMetadata.DigitalSignatureReference,
                        ProtectionScheme = row.ExchangeMetadata.ProtectionScheme,
                        DigitalSignatureValue = new DigitalSignatureValueObject
                        {
                            DigitalSignature = row.ExchangeMetadata.DigitalSignature.DigitalSignatureValue,
                            PublicCertificate = row.ExchangeMetadata.DigitalSignature.PublicCertificate,
                            PublicRootCertificateThumbprint = row.ExchangeMetadata.DigitalSignature.PublicRootCertificateThumbprint
                        }
                    },
                    AckRequest = (Common.Models.Public.AckRequestEnum)row.AckRequest
                };

                result.Add(obj);
            }
            return result;
        }

        private async Task<List<SummaryObject>> GetPublishedMessagesSummary(IQueryable<DataAccess.Entities.PublishedMessage> query, int page, int pageSize)
        {
            var result = new List<SummaryObject>();

            var data = await query.OrderBy(x => x.CreatedTime).Skip((page - 1) * pageSize).Take(pageSize).ToListAsync();

            foreach (var row in data)
            {
                var dt = _dbContext.DataTypeEnum.Where(x => x.Id == row.DataType).FirstOrDefault().Id;
                var obj = new SummaryObject
                {
                    DataReference = row.DataReference.ToString(),
                    DataType = (Common.Models.Public.DataTypeEnum)dt,
                    DataCompression = row.ExchangeMetadata.CompressionFlag,
                    DataProtection = row.ExchangeMetadata.DataProtection,
                    info_size = row.Data.Length,
                };
                result.Add(obj);
            }

            return result;
        }
        #endregion

    }
}
