﻿using System;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Logging;
using SECOM.Common.Models.Public;
using SECOM.DataAccess;
using SECOM.DataAccess.Entities;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/subscription")]
    [ApiController]
    public class SubscriptionController : ControllerBase
    {
        private SecomDbContext _dbContext;
        private readonly ILogger _logger;

        public SubscriptionController(SecomDbContext dbContext, ILogger<SubscriptionController> logger)
        {
            _dbContext = dbContext;
            _logger = logger;
        }

        /// <summary>
        /// The purpose of the interface is to request subscription on information, either specific information according to parameters, or the information accessible upon decision by information provider.
        /// </summary>
        /// <param name="request">Specific id on the information object subscription is requested for specific status on the information or specific area of interest</param>
        /// <returns>Confirmation or error message. Subscription identifier in return, if ok.</returns>
        /// <response code="200">Subscription successfully created</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>         
        /// <response code="403">Not authorized</response>
        /// <response code="404">Information not found</response>
        /// <response code="405">Method not allowed</response>        
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpPost]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(SubscriptionResponsObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]

        public async Task<ActionResult<SubscriptionResponsObject>> Post([FromBody] SubscriptionRequestObject request)
        {
            try
            {
                bool authorized = true;

                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new ResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}"});
                }

                if(!authorized)
                {
                    return Forbid($"Not authorized to requested information");
        
                }
                return Ok(new SubscriptionResponsObject
                {
                    Message = $"Subscription successfully created",
                    SubscriptionIdentifier = Guid.NewGuid()
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new UploadResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        /// <summary>
        /// This interface shall be used by the consumer to request removal of subscription.
        /// </summary>
        /// <param name="removeSubscriptionRequestObject">Specific identity of the information object to remove subscription for. If no id entity provided, all subscriptions for the caller is removed</param>
        /// <returns>Confirmation or error message</returns>
        /// <response code="200">Subscription *identifier* removed</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Not authorized to remove subscription</response>
        /// <response code="404">Subscription *identifier* not found</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>
        [HttpDelete]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<ResponseObject>> Delete([FromBody] RemoveSubscriptionObject removeSubscriptionRequestObject)
        {
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new ResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                var subscription = _dbContext.IdentityAcl.Where(x => x.Identity.Uid == removeSubscriptionRequestObject.SubscriptionIdentifier.ToString()).FirstOrDefault();
                if(subscription == null)
                {
                    return StatusCode(StatusCodes.Status404NotFound, new ResponseObject
                    {
                        Message = $"Subscription {removeSubscriptionRequestObject.SubscriptionIdentifier} not found."
                    });
                }

                return Ok(new ResponseObject
                {
                    Message = $"Subscription {removeSubscriptionRequestObject.SubscriptionIdentifier} removed."
                });
            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }


        /// <summary>
        /// The interface receives notifications when subscription is created or removed by information producer.
        /// </summary>
        /// <param name="subscriptionNotificationRequestObject">Contains the identity of the information object in focus and type of event; Create or Delete.</param>
        /// <returns>Confirmation or error message. Subscription identifier in return, if ok.</returns>
        /// <response code="200">Subscription *identifier* created/deleted notification received</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>  
        [HttpPost]
        [Route("notification")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(ResponseObject))]
        public async Task<ActionResult<ResponseObject>> Notification([FromBody] SubscriptionNotificationObject subscriptionNotificationRequestObject)
        {
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new ResponseObject
                    {
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                if(subscriptionNotificationRequestObject.EventEnum == SubscriptionEventEnum.Create)
                {
                    return Ok(new ResponseObject
                    {
                        Message = $"Subscription created notification for  id {subscriptionNotificationRequestObject.SubscriptionIdentifier} received successfully."
                    });
                }
                else
                {
                    return Ok(new ResponseObject
                    {
                        Message = $"Subscription deleted notification for id {subscriptionNotificationRequestObject.SubscriptionIdentifier} received successfully."
                    });
                }

            }
            catch (Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new UploadResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }
    }
}
