﻿using System;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using SECOM.Common.Models.Public;
using SECOM.Common.Schema;
using SECOM.Common.Services;
using SECOM.DataAccess;
using SECOM.DataAccess.Entities;
using SECOM.Service.Public.BackgroundTask;

namespace SECOM.Service.Public.Rest.v1
{
    [Route("v1/object")]
    [ApiController]
    public class UploadController : ControllerBase
    {
        private readonly ILogger _logger;
        private readonly IHttpContextAccessor _contextAccessor;
        private readonly ISignatureService _signatureVerificationService;
        private readonly IIdentityRegistryService _identityRegistryService;
        private readonly IServiceRegistryService _serviceRegistryService;
        private readonly ICertificateService _certificateService;
        private readonly IBackgroundTaskQueue _taskQueue;
        private readonly IServiceProvider _serviceProvider;

        private readonly SecomDbContext _dbContext;

        public UploadController(ILogger<UploadController> logger,
            IHttpContextAccessor contextAccessor,
            ISignatureService signatureVerificationService,
            IIdentityRegistryService identityRegistryService,
            IServiceRegistryService serviceRegistryService,
            ICertificateService certificateService,
            IBackgroundTaskQueue taskQueue,
            IServiceProvider serviceProvider,
            SecomDbContext dbContext)
        {
            _logger = logger;
            _contextAccessor = contextAccessor;
            _signatureVerificationService = signatureVerificationService;
            _dbContext = dbContext;
            _identityRegistryService = identityRegistryService;
            _serviceRegistryService = serviceRegistryService;
            _certificateService = certificateService;
            _taskQueue = taskQueue;
            _serviceProvider = serviceProvider;

            _logger.LogDebug($"{GetType().FullName} accessed by user {contextAccessor.HttpContext.User.Identity.Name} from {contextAccessor.HttpContext.Connection.RemoteIpAddress}");
        }

        /// <summary>
        /// The interface shall be used for uploading (pushing) data to a consumer. The operation expects one single data object and its metadata.
        /// </summary>
        /// <param name="uploadObject">The data (payload) package with its metadata</param>
        /// <returns>Confirmation of upload or error message, including description of incorrect or missing values in payload for the specific service.</returns>
        /// <response code="200">Message successfully uploaded</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Not authorized to upload</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="413">Request entity too large</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response>        
        [HttpPost]
        [Consumes("application/json")]
        [Produces("application/json")]
        [RequestSizeLimit(350000)]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status413RequestEntityTooLarge, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(UploadResponseObject))]

        public async Task<ActionResult<UploadResponseObject>> Post([FromBody] UploadObject uploadObject)
        {
            var response = string.Empty;

            try
            {
               // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService,
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                // Validate data type, only possible if not encrypted and/or inside an exchangeSet
                if(!uploadObject.Envelope.ExchangeMetadata.DataProtection && uploadObject.Envelope.ContainerType != Common.Models.Public.ContainerTypeEnum.S100_EXCHANGESET)
                {
                    try
                    {
                        //TODO: Add implementation specific schema validation
                        //var validator = new SchemaValidationService();
                        //TODO: Parse message from inside a dataset
                        //var data = Encoding.UTF8.GetString(uploadObject.Envelope.Data);

                        //validator.ValidateRtzXml(data);
                        //validator.ValidateS124Xml(data);
                        //etc.
                    }
                    catch (Secom_SchemaValidationException ex)
                    {
                        return BadRequest(new UploadResponseObject
                        {
                            Message = $"{ex.Message}",
                            SECOM_ResponseCode = Secom_ResponseCodeEnum.XmlSchemaValidationError
                        });
                    }
                }

                //Need to validate client certificate from TLS

                //Verify claimed public certificate in envelope against root thumbprint
                //TODO How do we verify the remaining certs, TLS, data 
                //Need thumbprint for verify TLS cert against e.g. Navelink
                var x509Cert = _certificateService.GetX509CertFromBase64(uploadObject.Envelope.EnvelopeSignatureCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, uploadObject.Envelope.EnvelopeRootCertificateThumbprint, out var errorMessage);
                response += errorMessage;
                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate,
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                //Check envelope signature
                var isValid = _signatureVerificationService.VerifyEnvelopeSignature(uploadObject.Envelope, uploadObject.EnvelopeSignature, 
                    uploadObject.Envelope.EnvelopeSignatureCertificate, out var message);
                response += message;

                if(!isValid)
                {
                    _logger.LogError($"Failed to verify envelope signature for message id {uploadObject.Envelope.TransactionIdentifier}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.FailedSignatureVerification,
                        Message = $"{string.Join(", ", message)}"
                    });
                }

                var uploadedMessage = new UploadedMessage
                {
                    AckEndpoint = await GetCallerEndpoint(),
                    AckRequest = (int)uploadObject.Envelope.AckRequest,
                    Data = uploadObject.Envelope.Data,
                    ContainerType = (int)uploadObject.Envelope.ContainerType,
                    DataProductType = (int)uploadObject.Envelope.DataProductType,
                    ExchangeMetadata = new SECOM.DataAccess.Entities.ExchangeMetadata
                    {
                        CompressionFlag = uploadObject.Envelope.ExchangeMetadata.CompressionFlag,
                        DataProtection = uploadObject.Envelope.ExchangeMetadata.DataProtection,
                        ProtectionScheme = uploadObject.Envelope.ExchangeMetadata.ProtectionScheme,
                        DigitalSignatureReference = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureReference,
                        DigitalSignature = new DigitalSignature
                        {
                            DigitalSignatureValue = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureValue.DigitalSignature,
                            PublicCertificate = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureValue.PublicCertificate,
                            PublicRootCertificateThumbprint = uploadObject.Envelope.ExchangeMetadata.DigitalSignatureValue.PublicRootCertificateThumbprint
                        },
                    },
                    FromSubscription = uploadObject.Envelope.FromSubscription,
                    TransactionIdentifier = uploadObject.Envelope.TransactionIdentifier,
                    ReceivedTime = DateTime.UtcNow,
                    EnvelopeSignatureCertificate = uploadObject.Envelope.EnvelopeSignatureCertificate,
                    EnvelopeRootCertificateThumbprint = uploadObject.Envelope.EnvelopeRootCertificateThumbprint,
                    EnvelopeSignature = uploadObject.EnvelopeSignature,
                    SignatureValidatedOk = isValid,
                    EnvelopeSignatureTime = uploadObject.Envelope.EnvelopeSignatureTime
                };

                await _dbContext.AddAsync(uploadedMessage);
                await _dbContext.SaveChangesAsync();

                _logger.LogInformation($"Successfully added uploaded message {uploadedMessage.TransactionIdentifier} to database.");
                return Ok(new UploadResponseObject
                {
                    Message = "Message successfully uploaded"
                });
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, ex.Message);
                if(ex.InnerException != null) _logger.LogError($"{ex.InnerException}");
                if(!string.IsNullOrEmpty(ex.StackTrace)) _logger.LogError(ex.StackTrace);
                return StatusCode(StatusCodes.Status500InternalServerError, new UploadResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }


        /// <summary>
        /// The purpose with this interface is to upload (push) a link to information to a consumer. 
        /// </summary>
        /// <param name="uploadLinkObject">The message link with its metadata</param>
        /// <returns>Confirmation of upload or error message</returns>
        /// <response code="200">Link successfully uploaded</response>
        /// <response code="400">Bad request</response>
        /// <response code="401">Unauthorized</response> 
        /// <response code="403">Not authorized to upload link</response>
        /// <response code="405">Method not allowed</response>
        /// <response code="500">Internal server error</response>
        /// <response code="501">Not implemented</response> 	
        [HttpPost]
        [Route("link")]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status401Unauthorized, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status403Forbidden, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status405MethodNotAllowed, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(UploadResponseObject))]
        [ProducesResponseType(StatusCodes.Status501NotImplemented, Type = typeof(UploadResponseObject))]
        public async Task<ActionResult<UploadResponseObject>> PostLink([FromBody] UploadLinkObject uploadLinkObject)
        {
            var response = string.Empty;
            try
            {
                // Validate incoming data
                if (!ModelState.IsValid)
                {
                    var errorList = ModelState.Values.SelectMany(m => m.Errors)
                                 .Select(e => e.ErrorMessage)
                                 .ToList();

                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.MissingRequiredDataForTheService,
                        Message = $"{string.Join(", ", errorList)}"
                    });
                }

                // Get caller url
                var sslCert = _contextAccessor.HttpContext.Connection.ClientCertificate;
                _logger.LogDebug($"{sslCert.FriendlyName}, {sslCert.Thumbprint}, {sslCert.Subject}");
                var callerMrn = _certificateService.GetInstanceIdFromCert(sslCert);
                _logger.LogDebug(callerMrn);

                //var serviceInstance = await _serviceRegistryService.FindService($"instanceId:{callerMrn.Replace(":", "\\:")}", "", 0, 1000);

                //Verify claimed public certificate in envelope against root thumbprint
                //TODO How do we verify the remaining certs, TLS, data 
                //Need thumbprint for verify TLS cert against e.g. Navelink
                var x509Cert = _certificateService.GetX509CertFromBase64(uploadLinkObject.Envelope.EnvelopeSignatureCertificate);
                var isValidCert = _certificateService.ValidateCertificate(x509Cert, uploadLinkObject.Envelope.EnvelopeRootCertificateThumbprint, out var errorMessage);

                response += errorMessage;
                if (!isValidCert)
                {
                    _logger.LogError($"Failed to get root certificate from {x509Cert.FriendlyName}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.InvalidCertificate,
                        Message = $"{string.Join(", ", errorMessage)}"
                    });
                }

                // Check envelope signature
                var isValid = _signatureVerificationService.VerifyEnvelopeSignature(uploadLinkObject.Envelope, uploadLinkObject.EnvelopeSignature,
                    uploadLinkObject.Envelope.EnvelopeSignatureCertificate, out var message);
                response += message;

                if (!isValid)
                {
                    _logger.LogError($"Failed to verify envelope signature for message id {uploadLinkObject.Envelope.TransactionIdentifier}.");
                    return BadRequest(new UploadResponseObject
                    {
                        SECOM_ResponseCode = Secom_ResponseCodeEnum.FailedSignatureVerification,
                        Message = $"{string.Join(", ", message)}"
                    });
                }

                var workItem = _serviceProvider.GetRequiredService<UploadLinkWorker>();

                //workItem.ContainerType = uploadLinkObject.Envelope.ContainerType;
                workItem.ClientCertificate = _contextAccessor.HttpContext.Connection.ClientCertificate;
                workItem.UploadLinkObject = uploadLinkObject;
                workItem.AckEndpoint = await GetCallerEndpoint();
                if(_contextAccessor.HttpContext.Request.Host.Host.ToLower().Contains("localhost"))
                {
                    var urls = Environment.GetEnvironmentVariable("ASPNETCORE_URLS")?.Split(";");
                    workItem.Url = $"{urls[0]}/v1/object/link?TransactionIdentifier={uploadLinkObject.Envelope.TransactionIdentifier}";
                }
                else
                {
                    workItem.Url = string.Empty;
                }
                _taskQueue.QueueBackgroundWorkItem(workItem.GetLink);
 
                return Ok(new UploadResponseObject
                {
                    Message = $"Link successfully uploaded with transactionId: {uploadLinkObject.Envelope.TransactionIdentifier}"
                });
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, ex.Message);
                return StatusCode(StatusCodes.Status500InternalServerError, new UploadResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        private async Task<string> GetCallerEndpoint()
        {
            string endPoint = "";
            // Get caller url
            var callerMrn = _certificateService.GetInstanceIdFromCert(_contextAccessor.HttpContext.Connection.ClientCertificate);
            var serviceInstance = await _serviceRegistryService.FindService($"instanceId:{callerMrn.Replace(":", "\\:")}", "", 0, 1000);

            if (serviceInstance != null && serviceInstance.Count == 1)
                endPoint = $"{serviceInstance.FirstOrDefault().EndpointUri}/v1/Acknowledgement/";

            _logger.LogInformation($"Stored caller acknowledgement endpoint: {endPoint}.");
            return endPoint;
        }

        private bool IsValidUri(string claimedUrl)
        {
            Uri uriResult;
            return Uri.TryCreate(claimedUrl, UriKind.Absolute, out uriResult) && uriResult.Scheme == Uri.UriSchemeHttp || uriResult.Scheme == Uri.UriSchemeHttps;
        }
    }
}
