using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Claims;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication.Certificate;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.OpenApi.Models;
using SECOM.Common.Services;
using SECOM.DataAccess;
using AutoMapper;
using SECOM.Service.Public.Filters;
using SECOM.Service.Public.BackgroundTask;

namespace SECOM.Service.Public
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        public void ConfigureServices(IServiceCollection services)
        {
            services.AddScoped<ICertificateService, CertificateService>();
            services.AddScoped<ISignatureService, SignatureService>();
            services.AddScoped<IIdentityRegistryService, IdentityRegistryService>();
            services.AddScoped<IServiceRegistryService, ServiceRegistryService>();
            services.AddScoped<IConvertService, ConvertService>();
            services.AddScoped<ICallService, CallService>();
            services.AddTransient<IEncryptionService>(p => new EncryptionService(256));
            //services.AddScoped<IEncryptionService, EncryptionService>();
            services.AddScoped<UploadLinkWorker, UploadLinkWorker>();
            services.AddScoped<EncryptionKeyWorker, EncryptionKeyWorker>();

            services.AddHostedService<QueuedHostedService>();
            services.AddSingleton<IBackgroundTaskQueue, BackgroundTaskQueue>();

            services.AddHttpContextAccessor();

            services.AddAuthentication(
                CertificateAuthenticationDefaults.AuthenticationScheme)
                .AddCertificate(options =>
                {
                    options.AllowedCertificateTypes = CertificateTypes.All;
                    options.RevocationMode = X509RevocationMode.NoCheck;
                    options.Events = new CertificateAuthenticationEvents
                    {
                        OnCertificateValidated = context =>
                        {
                            var validationService =
                                context.HttpContext.RequestServices
                                .GetService<ICertificateService>();

                            string rootThumbprint = Configuration.GetValue<string>("RootCertificateThumbprint");

                            if (validationService.ValidateCertificate(
                                context.ClientCertificate, rootThumbprint, out string errors))
                            {
                                // Extract data from certifcate
                                var certData = context.ClientCertificate.Subject.Split(',');

                                var id = string.Empty;
                                var name = string.Empty;
                                var orgId = string.Empty;

                                var certDataDictionary = new Dictionary<string, string>();
                                foreach (var item in certData)
                                {
                                    var parts = item.Split('=');
                                    if (parts != null && parts.Count() == 2)
                                    {
                                        if (parts[0].Trim().StartsWith("userId"))
                                            id = parts[1]; ;
                                        if (parts[0].Trim() == "O")
                                            orgId = parts[1]; ;
                                        if (parts[0].Trim() == "CN")
                                            name = parts[1]; ;

                                        certDataDictionary.Add(parts[0].Trim(), parts[1].Trim());
                                    }
                                }

                                var claims = new[]
                                {
                                    new Claim(
                                        ClaimTypes.NameIdentifier,
                                        id,
                                        ClaimValueTypes.String,
                                        context.Options.ClaimsIssuer),
                                    new Claim(
                                        ClaimTypes.Name,
                                        name,
                                        ClaimValueTypes.String,
                                        context.Options.ClaimsIssuer),
                                    new Claim(
                                        Constants.OrgClaimsType,
                                        orgId,
                                        ClaimValueTypes.String,
                                        context.Options.ClaimsIssuer)
                                };

                                context.Principal = new ClaimsPrincipal(
                                    new ClaimsIdentity(claims, context.Scheme.Name));
                                context.Success();
                            }
                            else
                            {
                                throw new UnauthorizedAccessException($"Invalid certificate{errors}");
                            }

                            return Task.CompletedTask;
                        },
                        OnAuthenticationFailed = context =>
                        {
                            return Task.CompletedTask;
                        },
                    };
                });

            var sqlConnectionString = Configuration.GetConnectionString("DataAccessPostgreSqlProvider");
            services.AddDbContext<SecomDbContext>(options =>
                    options.UseNpgsql(
                        sqlConnectionString,
                        b => b.MigrationsAssembly("SECOM.DataAccess")
                    )
                );

            services.AddHttpClient("IdentityRegistry", c => GetDefaultHttpClient(c, Configuration.GetValue<string>("ServiceEndpoints:IdentityRegistry")))
                .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());
            services.AddHttpClient("ServiceRegistry", c => GetDefaultHttpClient(c, Configuration.GetValue<string>("ServiceEndpoints:ServiceRegistry")))
                .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());
            services.AddHttpClient("GetByLink", c => GetDefaultHttpClient(c, null))
                .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());

            services.AddControllers();

            services.AddMvc()
                .ConfigureApiBehaviorOptions(options =>
                {
                    options.SuppressModelStateInvalidFilter = true;
                });

            services.AddSwaggerGen(c =>
            {
                c.SwaggerDoc("v1", new OpenApiInfo
                {
                    Title = "SECOM Information Service interface",
                    Version = "v1",
                    Description = "Compliant with SECOM 2021 edition 2021-07-01-r1",
                    Contact = new OpenApiContact
                    {
                        Name = "Swedish Maritime Administration",
                        Url = new Uri("http://sjofartsverket.se")
                    },
                });
                c.IncludeXmlComments(Path.Combine(AppDomain.CurrentDomain.BaseDirectory, @"SECOM.Service.Public.xml"));
            });

            services.AddAutoMapper(typeof(Startup));

            services.AddMvc(options => options.Filters.Add(new
                HandleExceptionAttribute()))
                .ConfigureApiBehaviorOptions(options =>
                {
                    options.SuppressModelStateInvalidFilter = true;
                });

        }

        public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
        {
            //app.UseMiddleware<ExceptionMiddleware>();

            app.UseAuthentication();

            app.UseSwagger();
            app.UseSwaggerUI(c =>
            {
                c.SwaggerEndpoint("./swagger/v1/swagger.json", "SECOM Information Service interface");
                c.DocumentTitle = "SECOM Service interface";
                c.RoutePrefix = string.Empty;
            });

            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }

            app.UseHttpsRedirection();

            app.UseRouting();

            app.UseAuthorization();

            app.UseEndpoints(endpoints =>
            {
                endpoints.MapControllers();
            });

            using (var scope = app.ApplicationServices.GetService<IServiceScopeFactory>().CreateScope())
            {
                scope.ServiceProvider.GetRequiredService<SecomDbContext>().Database.Migrate();
            }
        }

        private HttpClient GetDefaultHttpClient(HttpClient c, string baseUtl)
        {
            if (baseUtl != null)
                c.BaseAddress = new Uri(baseUtl);

            c.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
            c.DefaultRequestHeaders.CacheControl = new CacheControlHeaderValue
            {
                NoCache = true,
                NoStore = true,
                MaxAge = new TimeSpan(0),
                MustRevalidate = true
            };

            return c;
        }

        private HttpMessageHandler GetDefaultHttpMessageHandler()
        {
            var path = Configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Path");
            var password = Configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Password");

            var sslCert = new X509Certificate2(path, password);
            var httpHandler = new HttpClientHandler
            {
                ClientCertificateOptions = ClientCertificateOption.Manual,
                ServerCertificateCustomValidationCallback = HttpClientHandler.DangerousAcceptAnyServerCertificateValidator,
                SslProtocols =
                            System.Security.Authentication.SslProtocols.Tls12 |
                            System.Security.Authentication.SslProtocols.Tls11 |
                            System.Security.Authentication.SslProtocols.Tls

            };
            httpHandler.ClientCertificates.Add(sslCert);
            return httpHandler;
        }
    }
}
