﻿using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using SECOM.ServiceDiscovery.Models;
using SECOM.ServiceDiscovery.Services;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Web;

namespace SECOM.ServiceDiscovery.Rest.v1
{
    [Route("v1/[controller]")]
    [ApiController]
    public class SearchServiceController : ControllerBase
    {
        private const string PATH_SEARCH_GENERIC = "/api/_search/serviceInstance?query=";
        private const string PATH_SEARCH_WKT = "/api/_searchGeometryWKT/serviceInstance?geometry=[GEOMETRY]&query=";
        private const string PATH_ALL = "/api/serviceInstance";

        //private string _serviceRegistryBasePath;

        private IServiceRegistryService _service;


        public SearchServiceController(IConfiguration configuration, IServiceRegistryService serviceRegistryService)
        {
            //_serviceRegistryBasePath = configuration.GetValue<string>("ASPNETCORE_Kestrel:BasePath");
            _service = serviceRegistryService;
        }
        /// <summary>
        /// The purpose with this interface is to search for service instances to consume.
        /// </summary>
        /// <param name="searchObject ">The search request object</param>
        /// <returns>A success or an error message</returns>
        /// <response code="200">The operation was performed successfully</response>
        /// <response code="400">Bad request</response>
        /// <response code="404">Information not found</response>
        /// <response code="500">Internal server error</response>

        [HttpPost]
        [Consumes("application/json")]
        [Produces("application/json")]
        [ProducesResponseType(StatusCodes.Status200OK, Type = typeof(SearchServiceResponse))]
        [ProducesResponseType(StatusCodes.Status400BadRequest, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status404NotFound, Type = typeof(ResponseObject))]
        [ProducesResponseType(StatusCodes.Status500InternalServerError, Type = typeof(ResponseObject))]
        public async Task<ActionResult<SearchServiceResponse>> Post([FromBody] SearchServiceRequest searchObject)
        {
            string url = string.Empty;
            if(searchObject == null)
            {
                return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                {
                    Message = "Search object empty."
                });
            }
            var isFreeText = !string.IsNullOrEmpty(searchObject.FreeText);
            var isGeometry = !string.IsNullOrEmpty(searchObject.Geometry);

            try
            {
                if (isGeometry)
                {
                    url = PATH_SEARCH_WKT.Replace("[GEOMETRY]", HttpUtility.UrlEncode(searchObject.Geometry));
                }
                else
                {
                    url = PATH_SEARCH_GENERIC;
                }

                string query = string.Empty;
                if (isFreeText)
                {
                    if (FormatValidator.IsValidFreeText(searchObject.FreeText))
                    {
                        query = searchObject.FreeText.Replace(":", "\\:");
                        query = HttpUtility.UrlEncode(query);
                    }
                    else
                    {
                        string msg = "Forbidden character(s) in freetext search string.";
                        return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                        {
                            Message = msg
                        });
                    }
                }
                else
                {
                    if (!string.IsNullOrEmpty(searchObject.Query.Name))
                    {
                        query = AddToQuery(query, "name", searchObject.Query.Name, "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.Status))
                    {
                        query = AddToQuery(query, "status", searchObject.Query.Status, "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.Version))
                    {
                        query = AddToQuery(query, "version", searchObject.Query.Version, "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.Description))
                    {
                        query = AddToQuery(query, "description", searchObject.Query.Description, "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.SpecificationId))
                    {
                        query = AddToQuery(query, "specificationId", searchObject.Query.SpecificationId.Replace(":", "\\:"), "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.DesignId))
                    {
                        query = AddToQuery(query, "designId", searchObject.Query.DesignId.Replace(":", "\\:"), "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.InstanceId))
                    {
                        query = AddToQuery(query, "instanceId", searchObject.Query.InstanceId.Replace(":", "\\:"), "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.ServiceType))
                    {
                        query = AddToQuery(query, "serviceType", searchObject.Query.ServiceType, "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.Unlocode))
                    {
                        if (FormatValidator.IsValidUnlocode(searchObject.Query.Unlocode))
                        {
                            query = AddToQuery(query, "unlocode", searchObject.Query.Unlocode, "AND");
                        }
                        else
                        {
                            string msg = "Invalid unlocode format in search string.";
                            return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                            {
                                Message = msg
                            });
                        }
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.EndpointUri))
                    {
                        if (Uri.IsWellFormedUriString(searchObject.Query.EndpointUri, UriKind.Absolute))
                            query = AddToQuery(query, "endpointUri", searchObject.Query.EndpointUri.Replace(":", "\\:").Replace("/", "\\/"), "AND");
                        else
                        {
                            string msg = "Invalid uri format in search string.";
                            return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                            {
                                Message = msg
                            });
                        }
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.DataProductType))
                    {
                        query = AddToQuery(query, "dataProductType", searchObject.Query.EndpointUri.Replace(":", "\\:").Replace("/", "\\/"), "AND");
                    }

                    if (!string.IsNullOrEmpty(searchObject.Query.Mmsi) && !string.IsNullOrEmpty(searchObject.Query.Imo))
                    {
                        if (query == string.Empty)
                            query += "(";
                        else
                            query += " AND (";

                        if (FormatValidator.IsValidMmsi(searchObject.Query.Mmsi))
                        {
                            query = AddToQuery(query, "mmsi", searchObject.Query.Mmsi, "OR");
                        }
                        else
                        {
                            string msg = "Invalid mmsi format in search string.";
                            return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                            {
                                Message = msg
                            });
                        }
                        if(FormatValidator.IsValidImo(searchObject.Query.Imo))
                        {
                            query = AddToQuery(query, "imo", searchObject.Query.Imo, "OR");
                        }
                        else
                        {
                            string msg = "Invalid imo format in search string.";
                            return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                            {
                                Message = msg
                            });
                        }
                        query += ")";
                    }
                    else
                    {
                        if (!string.IsNullOrEmpty(searchObject.Query.Mmsi))
                        {
                            if (FormatValidator.IsValidMmsi(searchObject.Query.Mmsi))
                            {
                                query = AddToQuery(query, "mmsi", searchObject.Query.Mmsi, "AND");
                            }
                            else
                            {
                                string msg = "Invalid mmsi format in search string.";
                                return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                                {
                                    Message = msg
                                });
                            }
                        }

                        if (!string.IsNullOrEmpty(searchObject.Query.Imo))
                        {
                            if (FormatValidator.IsValidImo(searchObject.Query.Imo))
                            {
                                query = AddToQuery(query, "imo", searchObject.Query.Imo, "AND");
                            }
                            else
                            {
                                string msg = "Invalid imo format in search string.";
                                return StatusCode(StatusCodes.Status400BadRequest, new ResponseObject
                                {
                                    Message = msg
                                });
                            }
                        }
                    }

                    if (searchObject.Query.Keywords != null && searchObject.Query.Keywords.Count > 0)
                    {
                        if (query == string.Empty)
                            query += "(";
                        else
                            query += " AND (";

                        foreach (var id in searchObject.Query.Keywords)
                        {
                            query = AddToQuery(query, "keywords", id.Replace(":", "\\:"), "AND");
                        }
                        query += ")";
                    }
                }

                if (string.IsNullOrEmpty(query) && !isGeometry)
                {
                    url = PATH_ALL;
                }

                if (searchObject.Query.Page != null)
                {
                    if (url == PATH_ALL)
                        query += "?page=" + searchObject.Query.Page.ToString();
                    else
                        query += "&page=" + searchObject.Query.Page.ToString();
                }
                if (searchObject.Query.PageSize != null)
                {
                    query += "&size=" + searchObject.Query.PageSize.ToString();
                }

                SearchServiceResponse result;
                if (searchObject.IncludeXml.HasValue && searchObject.IncludeXml.Value)
                {
                    result = await _service.FindService(url + query, true);
                }
                else
                {
                    result = await _service.FindService(url + query, false);
                }

                if (result != null && result.ServiceInstances != null && result.ServiceInstances.Count > 0)
                    return result;
                else
                    return NotFound();
            }
            catch(Exception ex)
            {
                return StatusCode(StatusCodes.Status500InternalServerError, new ResponseObject
                {
                    Message = $"{ex.Message}"
                });
            }
        }

        private string AddToQuery(string query, string key, string value, string op)
        {
            var sb = new StringBuilder(query);
            if (sb.Length > 0 && !query.EndsWith("("))
                sb.Append(string.Format(" {0} {1}:{2}", op, key, value));
            else
                sb.Append(string.Format("{0}:{1}", key, value));

            return sb.ToString();
        }
    }
}
