using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Microsoft.OpenApi.Models;
using SECOM.ServiceDiscovery.Services;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Net.Http.Headers;
using System.Security.Cryptography.X509Certificates;
using System.Threading.Tasks;

namespace SECOM.ServiceDiscovery
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        // For more information on how to configure your application, visit https://go.microsoft.com/fwlink/?LinkID=398940
        public void ConfigureServices(IServiceCollection services)
        {
            services.AddScoped<IServiceRegistryService, ServiceRegistryService>();
            services.AddControllers();
            services.AddMvc()
                .ConfigureApiBehaviorOptions(options =>
                {
                    options.SuppressModelStateInvalidFilter = true;
                });

            services.AddHttpClient("ServiceRegistry", c => GetDefaultHttpClient(c, Configuration.GetValue<string>("ServiceEndpoints:ServiceRegistry")))
    .ConfigurePrimaryHttpMessageHandler(() => GetDefaultHttpMessageHandler());

            services.AddSwaggerGen(c =>
            {
                c.SwaggerDoc("v1", new OpenApiInfo
                {
                    Title = "SECOM Service Discovery interface",
                    Version = "v1",
                    Description = "Compliant with SECOM 2021 edition 2021-07-01-r1",
                    Contact = new OpenApiContact
                    {
                        Name = "Swedish Maritime Administration",
                        Url = new Uri("http://sjofartsverket.se")
                    },
                });
                c.IncludeXmlComments(Path.Combine(AppDomain.CurrentDomain.BaseDirectory, @"SECOM.ServiceDiscovery.xml"));
            });
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
        {
            app.UseSwagger();
            app.UseSwaggerUI(c =>
            {
                c.SwaggerEndpoint("./swagger/v1/swagger.json", "SECOM Service Discovery interface");
                c.DocumentTitle = "SECOM Service Discovery interface";
                c.RoutePrefix = string.Empty;
            });

            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }

            app.UseHttpsRedirection();

            app.UseRouting();

            app.UseAuthorization();

            app.UseEndpoints(endpoints =>
            {
                endpoints.MapControllers();
            });
        }

        private HttpClient GetDefaultHttpClient(HttpClient c, string baseUtl)
        {
            if (baseUtl != null)
                c.BaseAddress = new Uri(baseUtl);

            c.DefaultRequestHeaders.Accept.Add(new MediaTypeWithQualityHeaderValue("application/json"));
            c.DefaultRequestHeaders.CacheControl = new CacheControlHeaderValue
            {
                NoCache = true,
                NoStore = true,
                MaxAge = new TimeSpan(0),
                MustRevalidate = true
            };

            return c;
        }

        private HttpMessageHandler GetDefaultHttpMessageHandler()
        {
            var path = Configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Path");
            var password = Configuration.GetValue<string>("ASPNETCORE_Kestrel:Certificates:Default:Password");

            var sslCert = new X509Certificate2(path, password);
            var httpHandler = new HttpClientHandler
            {
                ClientCertificateOptions = ClientCertificateOption.Manual,
                ServerCertificateCustomValidationCallback = HttpClientHandler.DangerousAcceptAnyServerCertificateValidator,
                SslProtocols =
                            System.Security.Authentication.SslProtocols.Tls12 |
                            System.Security.Authentication.SslProtocols.Tls11 |
                            System.Security.Authentication.SslProtocols.Tls

            };
            httpHandler.ClientCertificates.Add(sslCert);
            return httpHandler;
        }
    }
}
